
# include "semesh_mesh.h"

bool ReadMap=true;

//============================== Vertex ========================

SrOutput& operator<< ( SrOutput& out, const Vertex& v )
 {
   return out << v.p << srspc << v.c;
 }

SrInput& operator>> ( SrInput& inp, Vertex& v )
 {
   return inp >> v.p >> v.c;
 }

//============================== Edge ========================

SrOutput& operator<< ( SrOutput& out, const Edge& e )
 {
   return out << e.c;
 }

SrInput& operator>> ( SrInput& inp, Edge& e )
 {
   return inp >> e.c;
 }

//============================== Face ========================

SrOutput& operator<< ( SrOutput& out, const Face& f )
 {
   return out << f.c;
 }

SrInput& operator>> ( SrInput& inp, Face& f )
 {
   return inp >> f.c;
 }

//============================== Mesh ========================

SymEdge *Mesh::init ( const SrPnt &p1, const SrPnt &p2 ) 
 { 
   SymEdge *s = MeshBase::init();
   s->vtx()->p = p1;
   s->sym()->vtx()->p = p2;
   return s;
 }

SymEdge *Mesh::mev ( SymEdge *s, const SrPnt &p )
 { 
   s = MeshBase::mev(s);
   if ( !s ) return s;
   s->vtx()->p = p;
   return s;
 }

SymEdge *Mesh::mev ( SymEdge *s1, SymEdge *s2, const SrPnt &p ) 
 { 
   SymEdge *s = MeshBase::mev(s1,s2); 
   if ( !s ) return s;
   s->vtx()->p = p;
   return s;
 }

SymEdge *Mesh::vsplit ( SymEdge *s1, SymEdge *s2, const SrPnt &p ) 
 { 
   SymEdge *s = MeshBase::vsplit(s1,s2); 
   if ( !s ) return s;
   s->vtx()->p = p;
   return s;
 }

SymEdge *Mesh::addv ( SymEdge *x, const SrPnt &p  )
 {
   SymEdge* s = MeshBase::addv(x);
   if ( s ) s->vtx()->p = p;
   return s;
 }

void Mesh::face_to_polygon ( Face* f, SrPolygon& pol )
 {
   SymEdge *s, *si;
   s = si = f->se();
   pol.capacity ( vertices_in_face(s) );
   pol.size(0);
   do { pol.push() = s->vtx()->p;
        s = s->nxt();
      } while ( s!=si );
 } 


//============================== EOF ========================

