
# ifndef SRMESH_MESH_H
# define SRMESH_MESH_H

# include <SR/sr_vec.h>
# include <SR/sr_color.h>
# include <SR/sr_polygon.h>
# include <SE/se_mesh.h>

//******************************************************************
//  First, we define the types we will use
//******************************************************************

class Vertex;
class Edge;
class Face;

typedef Se < Vertex, Edge, Face > SymEdge;

//******************************************************************
//  Here we define our vertex, edge and face types
//******************************************************************

class Vertex : public SeElement
 { public :
    SrPnt p;
    SrColor c;
   public :
    SE_ELEMENT_CASTED_METHODS(Vertex,SymEdge);
    Vertex () {}
    Vertex ( const Vertex& v ) { p=v.p; c=v.c; }
    friend SrOutput& operator<< ( SrOutput& out, const Vertex& v );
    friend SrInput& operator>> ( SrInput& inp, Vertex& v );
    friend int sr_compare ( const Vertex* v1, const Vertex* v2 ) { return 0; } // not used
 };

class Edge : public SeElement
 { public :
    SrColor c;
   public :
    SE_ELEMENT_CASTED_METHODS(Edge,SymEdge);
    Edge () {}
    Edge ( const Edge& e ) { c=e.c; }
    friend SrOutput& operator<< ( SrOutput& out, const Edge& e );
    friend SrInput& operator>> ( SrInput& inp, Edge& e );
    friend int sr_compare ( const Edge* e1, const Edge* e2 ) { return 0; } // not used
 };

class Face : public SeElement
 { public :
    SrColor c;
   public :
    SE_ELEMENT_CASTED_METHODS(Face,SymEdge);
    Face () {}
    Face ( const Face& f ) { c=f.c; }
    friend SrOutput& operator<< ( SrOutput& out, const Face& f );
    friend SrInput& operator>> ( SrInput& inp, Face& f );
    friend int sr_compare ( const Face* f1, const Face* f2 ) { return 0; } // not used
 };

//******************************************************************
//  Now we define our Mesh with some custom methods
//******************************************************************

typedef SeMesh <Vertex,Edge,Face> MeshBase;
class Mesh : public MeshBase
 { public :
    SymEdge* init ( const SrPnt& p1, const SrPnt& p2 );
    SymEdge* mev ( SymEdge* s, const SrPnt& p );
    SymEdge* mev ( SymEdge* s1, SymEdge* s2, const SrPnt& p );
    SymEdge* vsplit ( SymEdge* s1, SymEdge* s2, const SrPnt& p );
    SymEdge* addv ( SymEdge* x, const SrPnt& p );
    void face_to_polygon ( Face* f, SrPolygon& pol );
 };

# endif // SRMESH_MESH_H

