
# include <stdlib.h>

# include <FL/fl_ask.h>

# include <SR/sr_polygon.h>

# include "semesh_main_win.h"
# include "semesh_main.h"
# include "semesh_mesh.h"
# include "semesh_viewer.h"

Viewer::Viewer ( int x, int y, int w, int h ) : SrViewer(x,y,w,h) 
 {
   _vertices = new SrSnPoints;
   _edges = new SrSnLines;
   _faces = new SrSnGroup;
   _redsym = new SrSnLines;
   _bluesym = new SrSnLines;
   _point = new SrSnPoints;

   SrSnGroup* r = new SrSnGroup;
   r->add ( _faces );
   r->add ( _edges );
   r->add ( _vertices );
   r->add ( _bluesym );
   r->add ( _redsym );
   r->add ( _point );
   SrViewer::root ( r );

   menu_cmd ( CmdSpinAnim );
   menu_cmd ( CmdPlanar );
 }

void Viewer::visibility ( bool v, bool e, bool f, bool r, bool b, bool p )
 {
   _vertices->visible ( v );
   _edges->visible ( e );
   _faces->visible ( f );
   _redsym->visible ( r );
   _bluesym->visible ( b );
   _point->visible ( p );
 }

static void arrow ( SymEdge *e, SrVec &a, SrVec &b, SrVec &c )
 {
   a = e->vtx()->p;
   b = e->nxt()->vtx()->p;
   c = (a-b)*0.2f;
   SrVec ortho(c.y,-c.x,0);
   c = b + c + ortho;
 }

void Viewer::draw_parms ()
 {
   SrLines& red = _redsym->shape();
   SrLines& blu = _bluesym->shape();
   SrPoints& pnt = _point->shape();

   red.init ();
   blu.init ();
   pnt.init ();

   if ( TheMesh->empty() ) return;

   //=== blue ===
   SrPnt a, b, c;
   arrow ( TheBlueSymEdge, a, b, c );
   blu.push_color ( SrColor::blue );
   blu.begin_polyline ();
   blu.push_vertex ( a );
   blu.push_vertex ( b );
   blu.push_vertex ( c );
   blu.end_polyline ();

   //=== red ===
   arrow ( TheRedSymEdge, a, b, c );
   red.push_color ( SrColor::red );
   red.begin_polyline ();
   red.push_vertex ( a );
   red.push_vertex ( b );
   red.push_vertex ( c );
   red.end_polyline ();

   //=== point ===
   pnt.init_with_attributes ();
   pnt.push ( *ThePoint, SrColor::black, 4.0f );
 }

void Viewer::draw_mesh ()
 {
   SrPoints& vtx = _vertices->shape();
   SrLines& edg = _edges->shape();

   vtx.init ();
   edg.init ();
   _faces->remove_all ();

   if ( TheMesh->empty() ) return;

   //=== vertices ===
   Vertex *v, *vi;
   vtx.init_with_attributes ();
   v = vi = TheMesh->first()->vtx();
   do { vtx.push ( v->p, v->c, 3.0f );
        v = v->nxt();
      } while ( v!=vi );

   //=== edges ===
   Edge *e, *ei;
   e = ei = TheMesh->first()->edg();
   do { edg.push_color ( e->c );
        edg.push_line ( e->se()->vtx()->p, e->se()->nxt()->vtx()->p );
        e = e->nxt();
      } while ( e!=ei );

   //=== faces ===
   Face *f, *fi;
   SrPolygon pol;
   SrSnPolygon* spol;
   f = fi = TheMesh->first()->fac();
   do { TheMesh->face_to_polygon(f,pol);
        if ( pol.is_simple() && pol.is_ccw() )
         { spol = new SrSnPolygon;
           spol->color ( f->c );
           spol->shape() = pol;
           _faces->add ( spol );
	     }
        f = f->nxt();
      } while ( f!=fi );

/*

//-0.86196 0.95215
double end=123;
double Ex[]={
-0.82891,+0.77064,
-0.82353,+0.79908,
-0.82339,+0.80009,
-0.82316,+0.80110,
-0.82288,+0.80210,
-0.84006,+0.92464,
-0.84046,+0.92557,
-0.84080,+0.92654,
-0.84111,+0.92752,
-0.84140,+0.92849,
-0.84168,+0.92948,
-0.84194,+0.93047,
-0.84219,+0.93144,
-0.84244,+0.93242,
-0.84269,+0.93345,
-0.84293,+0.93448,
-0.84314,+0.93550,
-0.84334,+0.93653,
-0.84352,+0.93756,
-0.84368,+0.93859,
-0.84383,+0.93962,
-0.84395,+0.94065,
-0.84406,+0.94167,
-0.84415,+0.94270,
-0.84423,+0.94373,
-0.84428,+0.94476,
-0.84432,+0.94579,
-0.84433,+0.94681,
-0.84434,+0.94784,
-0.84434,+0.94886,
-0.84434,+0.94986,
-0.84434,+0.95091,
-0.84434,+0.95194,
-0.84434,+0.95296,
-0.84434,+0.95397,
-0.86147,+0.97279,
-0.86244,+0.97240,
-2.00000,+2.00000, end };

double* pt = Ex;
edg.init();
edg.begin_polyline();
while ( *pt!=end )
 { edg.push_vertex ( (float)pt[0], (float)pt[1] );
   pt+=2;
 }
edg.end_polyline();

vtx.init_with_attributes ();
vtx.push ( -0.86196f, 0.95215f, SrColor::red, 3.0f );
*/

 }

//==================== virtuals ===================================

int Viewer::handle_event ( const SrEvent &e )
 {
   return SrViewer::handle_event(e);
 }

int Viewer::handle_keyboard ( const SrEvent& e )
 {
   if ( e.key==27 )
    { exit(0); 
    }

   return SrViewer::handle_keyboard(e);
 }

int Viewer::handle_scene_event ( const SrEvent& e )
 {
   if ( e.type==SrEvent::Push || e.type==SrEvent::Drag )
    { *ThePoint = e.mousep;
      draw_parms();
      TheMainWin->render();

      return 1; // tell that I understood the event
    }

   return SrViewer::handle_scene_event(e);
 }
