
# ifndef CMESH_H
# define CMESH_H

//******************************************************************
// Here is an example of the types required to define a SeMesh
// containing 2d coordinates in the vertices, and user data in
// edges to store the ids of constrained edges.
//******************************************************************

//******************************************************************
//  First, we define the types we will use
//******************************************************************

class CmVertex; // will be defined later containing my own data
class CmEdge;   // will be defined later containing constraints ids
class CmFace;   // this will be a default element

// Define MySymEdge to have nice type casts defined:
typedef Se < CmVertex, CmEdge, CmFace > MySymEdge;

// Define MyMesh also to have nice type casts defined:
typedef SeMesh < CmVertex, CmEdge, CmFace > MyMesh;

//******************************************************************
//  Second, we define the face as a default element
//******************************************************************

SE_DEFINE_DEFAULT_ELEMENT(CmFace,MySymEdge);

//******************************************************************
//  Then we define the vertex class with the 2d coordinates
//******************************************************************

class CmVertex : public SeElement
 { public :
    float c[2]; // 2d coordinates
   public :
    SE_ELEMENT_CASTED_METHODS(CmVertex,MySymEdge);
    CmVertex () { c[0]=c[1]=0; }
    CmVertex ( const CmVertex& v ) { c[0]=v.c[0]; c[1]=v.c[1]; }

    void set ( float x, float y ) { c[0]=x; c[1]=y; }

    friend SrOutput& operator<< ( SrOutput& out, const CmVertex& v )
           { return out << v.c[0] << srspc << v.c[1]; }

    friend SrInput& operator>> ( SrInput& inp, CmVertex& v )
           { return inp >> v.c[0] >> v.c[1]; }

    friend int sr_compare ( const CmVertex* v1, const CmVertex* v2 ) { return 0; } // not used
 };

//******************************************************************
//  And finally we define the edge class with the ids
//******************************************************************

class CmEdge : public SeElement
 { public :
    SrArray<int> constraints_ids; // ids of all constraints sharing this edge
   public :
    SE_ELEMENT_CASTED_METHODS(CmEdge,MySymEdge);
    CmEdge () { }
    CmEdge ( const CmEdge& e ) { constraints_ids=e.constraints_ids; }

    bool is_constrained() { return constraints_ids.size()>0? true:false; }

    friend SrOutput& operator<< ( SrOutput& out, const CmEdge& e )
           { return out << e.constraints_ids; }

    friend SrInput& operator>> ( SrInput& inp, CmEdge& e )
           { return inp >> e.constraints_ids; }

    friend int sr_compare ( const CmEdge* e1, const CmEdge* e2 ) { return 0; } // not used
 };

# endif // CMESH_H
