
# include <stdlib.h>
# include <SE/se_mesh.h>
# include <SE/se_mesh_import.h>

//******************************************************************
//  Used Mesh Types
//******************************************************************

# include "setut_vmesh.h"
typedef VmVertex MyVertex;
typedef VmEdge   MyEdge;
typedef VmFace   MyFace;

//******************************************************************
//  Example Data to Import
//******************************************************************

static float Vertices[] = 
 { -0.5,  0.0,
    0.0, -0.5,
    0.0,  0.5,
    0.5,  0.0  };

static int Triangles[] =
 { 0, 1, 2,
   2, 3, 1 };

static int NumTriangles = 2;
static int NumVertices = 4;

//******************************************************************
//  Definition of MyImport class
//******************************************************************

class MyImport : public SeMeshImport
 { public :
    virtual SeMeshBase* get_new_shell ();
    virtual void attach_vtx_info ( SeVertex* v, int vi );
 };

SeMeshBase* MyImport::get_new_shell ()
 {
   return new MyMesh;
 }

void MyImport::attach_vtx_info ( SeVertex* v, int vi )
 {
   MyVertex* mv = (MyVertex*)v;
   mv->c[0] = Vertices[vi*2];
   mv->c[1] = Vertices[vi*2+1];
 }

//******************************************************************
//  Draw the mesh with fltk/opengl
//******************************************************************

# include <FL/Fl_Gl_Window.H>
# include <FL/gl.h>
# include <FL/Fl.H>

static MyMesh* TheMesh; // a global pointer to our mesh, to make life easier

class ImpViewer : public Fl_Gl_Window 
 { public : 
    ImpViewer ( int x, int y, int w, int h, const char *label=0 )
         :Fl_Gl_Window ( x, y, w, h, label )
    { Fl::gl_visual ( FL_RGB8 | FL_DOUBLE | FL_DEPTH );
    }

    virtual void draw ();
 };

void ImpViewer::draw ()
 {
   MySymEdge *s;
   MyEdge *e, *ei;

   glEnable ( GL_LINE_SMOOTH );
   glHint ( GL_LINE_SMOOTH_HINT, GL_NICEST );
   glClearColor ( 1, 1, 1, 0 );

   if ( TheMesh->empty() ) return;

   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   // draw edges:
   glLineWidth ( 0.5f );
   glBegin ( GL_LINES );
   glColor3f ( 0, 0, 0.8f );
     
   // loop over the circular list of edges:
   e = ei = TheMesh->first()->edg();
   do { s = e->se();
        glVertex2fv ( s->vtx()->c );
        glVertex2fv ( s->nxt()->vtx()->c );
        e = e->nxt();
      } while ( e!=ei );

   glEnd ();
 }

//******************************************************************
//  The main function
//******************************************************************

void import_main ()
 {
   // Init MyImport:
   MyImport* imp = new MyImport;

   // Do the import:
   MyImport::Msg msg = imp->start ( Triangles, NumTriangles, NumVertices );
   while ( msg!=MyImport::MsgFinished )
    { msg = imp->next_step ();
    }

   // Get result:
   if ( imp->shells.size()!=1 )
    { printf("Error: shell Not Found!\n"); exit(1); }

   TheMesh = (MyMesh*)imp->shells.pop();
   printf ( "Mesh found:\n" );
   printf ( "vertices: %d\n" , TheMesh->vertices() );
   printf ( "edges: %d\n" , TheMesh->edges() );
   printf ( "faces: %d\n" , TheMesh->faces() );

   // Initialize fltk and show the window:
   ImpViewer* viewer = new ImpViewer ( 200, 200, 400, 300 );
   viewer->label ( "SeMesh Import Example - M. Kallmann 2003" );
   viewer->show();

   Fl::run();
 }
