    
# include <stdlib.h>
# include <SE/se_mesh.h>

//******************************************************************
//  Used Mesh Types
//******************************************************************

# include "setut_vmesh.h"
typedef VmVertex MyVertex;
typedef VmEdge   MyEdge;
typedef VmFace   MyFace;

//******************************************************************
// Ok, now we can use the mesh and draw it with OpenGL (and fltk)
//******************************************************************

# include <FL/Fl_Gl_Window.H>
# include <FL/gl.h>
# include <FL/Fl.H>

static MyMesh* TheMesh; // a global pointer to our mesh, to make life easier

class MeshViewer : public Fl_Gl_Window 
 { public : 
    MeshViewer ( int x, int y, int w, int h, const char *label=0 )
         :Fl_Gl_Window ( x, y, w, h, label )
    { Fl::gl_visual ( FL_RGB8 | FL_DOUBLE | FL_DEPTH );
    }

    virtual void draw ();
 };

void MeshViewer::draw ()
 {
   MySymEdge *s;
   MyEdge *e, *ei;

   glEnable ( GL_LINE_SMOOTH );
   glHint ( GL_LINE_SMOOTH_HINT, GL_NICEST );
   glClearColor ( 1, 1, 1, 0 );

   if ( TheMesh->empty() ) return;

   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   // draw edges:
   glLineWidth ( 0.5f );
   glBegin ( GL_LINES );
   glColor3f ( 0, 0, 0.8f );
     
   // loop over the circular list of edges:
   e = ei = TheMesh->first()->edg();
   do { s = e->se();
        glVertex2fv ( s->vtx()->c );
        glVertex2fv ( s->nxt()->vtx()->c );
        e = e->nxt();
      } while ( e!=ei );

   glEnd ();
 }

//******************************************************************
// The Main Function
//******************************************************************

void mesh_main ()
 {
   float xmin, xmax, ymin, ymax;

   // Create the mesh object we have defined:
   TheMesh = new MyMesh;

   // Construct a simple planar mesh with two triangles and a square border:
   xmin = ymin = -0.8f;
   xmax = ymax =  0.8f;
   MySymEdge* s = TheMesh->init();
   s->vtx()->set ( xmax, ymax );
   s->nxt()->vtx()->set ( xmin, ymax );
   s = TheMesh->mev ( s );
   s->vtx()->set ( xmax, ymin );
   s= TheMesh->mev ( s );
   s->vtx()->set ( xmin, ymin );
   TheMesh->mef ( s, s->nxt()->nxt()->nxt() ); // close the square
   TheMesh->mef ( s, s->nxt()->nxt() ); // triangulate square

   // print it to the console:
   sr_out<<srnl;
   TheMesh->save ( sr_out );

   // Initialize fltk and show the window:
   MeshViewer* viewer = new MeshViewer ( 200, 200, 400, 300 );
   viewer->label ( "SeMesh Example - M. Kallmann 2003" );
   viewer->show();

   Fl::run();
 }

/* You should see in your shell the folowing mesh description:

SYMEDGE MESH DESCRIPTION

SymEdges 10
3 6 1 0 2
6 8 0 0 0
8 5 2 1 1
5 1 0 1 2
2 9 3 2 1
7 2 2 2 2
9 0 1 3 0
0 4 3 3 2
4 3 0 4 1
1 7 3 4 0

Vertices 4
1 0.800000 0.800000
0 -0.800000 0.800000
2 0.800000 -0.800000
4 -0.800000 -0.800000

Edges 5
0 
2 
4 
6 
8 

Faces 3
9 
8 
7 
*/
