
# include <stdlib.h>
# include <SE/se_mesh.h>
# include <SE/se_triangulator.h>

//******************************************************************
//  Used Mesh Types
//******************************************************************

# include "setut_vmesh.h"
typedef VmVertex MyVertex;
typedef VmEdge   MyEdge;
typedef VmFace   MyFace;

//******************************************************************
//  My Triangulator Manager
//******************************************************************

class MyTriFaceManager: public SeTriangulatorManager
 { public :
    virtual void get_vertex_coordinates ( const SeVertex* v, double& x, double & y );
    virtual void set_vertex_coordinates ( SeVertex* v, double x, double y );
 };

void MyTriFaceManager::get_vertex_coordinates ( const SeVertex* v, double& x, double & y )
 {
   x = (double) ((MyVertex*)v)->c[0];
   y = (double) ((MyVertex*)v)->c[1];
 }

void MyTriFaceManager::set_vertex_coordinates ( SeVertex* v, double x, double y )
 {
   ((MyVertex*)v)->c[0] = (float) x;
   ((MyVertex*)v)->c[1] = (float) y;
 }

//******************************************************************
//  Ready to triangulate a face
//******************************************************************

# define POLYSIZE 9

float poly[] = { -0.5f, -0.7f,
                 -0.7f, -0.2f,
                 -0.8f, 0.4f,
                 -0.1f, 0.7f,
                  0.8f, 0.6f,
                  0.8f, 0.0f,
                  0.1f, 0.3f,
                 -0.3f, -0.4f,
                  0.8f, -0.4f }; // a non-convex polygon

static MyMesh* TheMesh; // a global pointer to our mesh, to make life easier

# include <FL/Fl_Gl_Window.H>
# include <FL/gl.h>
# include <FL/Fl.H>

class TfViewer : public Fl_Gl_Window 
 { public : 
    TfViewer ( int x, int y, int w, int h, const char *label=0 )
         :Fl_Gl_Window ( x, y, w, h, label )
    { Fl::gl_visual ( FL_RGB8 | FL_DOUBLE | FL_DEPTH );
    }

    virtual void draw ();
 };

void TfViewer::draw ()
 {
   MySymEdge *s;
   MyEdge *e, *ei;

   glEnable ( GL_LINE_SMOOTH );
   glHint ( GL_LINE_SMOOTH_HINT, GL_NICEST );
   glClearColor ( 1, 1, 1, 0 );

   if ( TheMesh->empty() ) return;

   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   // draw edges:
   glLineWidth ( 0.5f );
   glBegin ( GL_LINES );
   glColor3f ( 0, 0, 0.8f );
     
   // loop over the circular list of edges:
   e = ei = TheMesh->first()->edg();
   do { s = e->se();
        glVertex2fv ( s->vtx()->c );
        glVertex2fv ( s->nxt()->vtx()->c );
        e = e->nxt();
      } while ( e!=ei );

   glEnd ();
 }

void triface_main ()
 {
   // Init used mesh and triangulator types:
   TheMesh = new MyMesh;
   SeTriangulator* tri = new SeTriangulator
                      ( SeTriangulator::ModeUnconstrained, // no constraints are used here
                        TheMesh,                           // my mesh to triangulate
                        new MyTriFaceManager,              // my manager says how to access my mesh
                        0.0001                             // the used epsilon in geometric primitives
                      );

   // Create the face and triangulate it:
   MySymEdge *s = TheMesh->init();
   MySymEdge *x = s->nxt();
   x->vtx()->set ( poly[0], poly[1] );
   s->vtx()->set ( poly[2], poly[3] );
   for ( int i=2; i<POLYSIZE; i++ )
    { s = TheMesh->mev ( s );
      s->vtx()->set ( poly[i*2], poly[i*2+1] );
    }
   TheMesh->mef ( s, x );

   // Attention: the face must be in ccw orientation!
   tri->triangulate_face ( s->fac() );

   // Initialize fltk and show the window:
   TfViewer* viewer = new TfViewer ( 200, 200, 400, 300 );
   viewer->label ( "SeMesh Face Triangulation Example - M. Kallmann 2003" );
   viewer->show();

   Fl::run();
 }
