
# ifndef VMESH_H
# define VMESH_H

//******************************************************************
// Here is an example of the types required to define a SeMesh
// containing only user data in the vertices (2d coordinates)
//******************************************************************

//******************************************************************
//  First, we define the types we will use
//******************************************************************

class VmVertex; // will be defined later containing my own data
class VmEdge;  // this will be a default element
class VmFace;  // this will be a default element

// Define MySymEdge to have nice type casts defined:
typedef Se < VmVertex, VmEdge, VmFace > MySymEdge;

// Define MyMesh also to have nice type casts defined:
typedef SeMesh < VmVertex, VmEdge, VmFace > MyMesh;

//******************************************************************
//  Second, we define the edge and face as default elements
//******************************************************************

SE_DEFINE_DEFAULT_ELEMENT(VmEdge,MySymEdge);
SE_DEFINE_DEFAULT_ELEMENT(VmFace,MySymEdge);

//******************************************************************
//  Finally, we define the vertex class we want to use
//******************************************************************

class VmVertex : public SeElement
 { public :
    float c[2]; // 2d coordinates
   public :
    SE_ELEMENT_CASTED_METHODS(VmVertex,MySymEdge);
    VmVertex () { c[0]=c[1]=0; }
    VmVertex ( const VmVertex& v ) { c[0]=v.c[0]; c[1]=v.c[1]; }

    void set ( float x, float y ) { c[0]=x; c[1]=y; }

    friend SrOutput& operator<< ( SrOutput& out, const VmVertex& v )
           { return out << v.c[0] << srspc << v.c[1]; }

    friend SrInput& operator>> ( SrInput& inp, VmVertex& v )
           { return inp >> v.c[0] >> v.c[1]; }

    friend int sr_compare ( const VmVertex* v1, const VmVertex* v2 ) { return 0; } // not used
 };

# endif // VMESH_H
