
# include <FL/Fl.H>
# include <FL/fl_draw.H>
# include <FL/Fl_Color_Chooser.H>
# include <FL/Fl_File_Chooser.H>

# include <SR/sr_fl.h>
# include <SR/sr_output_window.h>
# include <SR/sr_matrix_window.h>

# include "sr_fl_ui.h"

//================================ funcs =================================================

bool fl_color_chooser ( const char *title, SrColor& c )
 {
   int changed = fl_color_chooser(title,c.r,c.g,c.b);

   return changed? true:false;
 }

static UiInputWin* input_win ( const char* title, const char* label,
                        float* float_value, const char* string_value )
 {
   static UiInputWin* win = 0;
   if ( !win ) win = new UiInputWin;

   win->window->label ( title );
   win->label_box->label ( label );

   if ( float_value )
    { win->value_input->value ( (double)*float_value );
      win->text_input->hide ();
      win->value_input->show ();
    }
   else if ( string_value )
    { win->text_input->value ( string_value );
      win->text_input->show ();
      win->value_input->hide ();
    }

   win->window->show();
   while ( win->window->shown() ) Fl::check();

   return win;
 }

bool fl_value_input ( const char* title, const char* label, float& value )
 {
   UiInputWin* win = input_win ( title, label, &value, 0 );
   value = (float) win->value_input->value();
   return win->last_window_result;   
 }

bool fl_string_input ( const char* title, const char* label, SrString& s )
 {
   UiInputWin* win = input_win ( title, label, 0, s );
   s = win->text_input->value();
   return win->last_window_result;   
 }

void fl_text_output ( const char* title, const char* text )
 {
   static SrOutputWindow* win = 0;
   if ( !win ) win = new SrOutputWindow;

   win->title ( title );
   win->clear ();
   win->output ( text );
   win->show_first_line ();
   win->show ();
 }

bool fl_matrix_input ( const char* title, SrMat& m )
 {
   static SrMatrixWindow* win = 0;
   if ( !win )
    { win = new SrMatrixWindow;
      win->window()->set_modal();
    }

   if ( title ) win->title ( title );
   win->mat ( m );

   win->window()->show();

   while ( win->window()->shown() ) Fl::check();

   if ( win->ok_was_pressed() ) m = win->mat();

   return win->ok_was_pressed();
 }

void fl_scan_dir ( const SrString& dir,
                   SrStringArray& dirs,
                   SrStringArray& files,
                   const SrStringArray& ext )
 {
   dirent **list;

   int n = fl_filename_list ( dir, &list );
   int i;
   SrString s;

   for ( i=0; i<n; i++ )
    { if ( list[i]->d_name[0]=='.' ) continue; // "./" or "../"
      s = dir;
      s.append ( list[i]->d_name );
      if ( fl_filename_isdir(s) )
       { dirs.push ( s );
       }
      else // is file
       { int j;
         for ( j=0; j<ext.size(); j++ ) 
          if ( s.has_file_extension(ext[j]) ) break;
         if ( ext.size()==0 || j<ext.size() )
          files.push ( s );
       }
    }

   for ( i=0; i<n; i++ ) free ( (void*)(list[i]) );
   free((void*)list);
 }


void fl_scan_files ( const SrString& basedir, SrStringArray& files, 
                     const SrStringArray& ext )
 {
   SrStringArray dirs;

   fl_scan_dir ( basedir, dirs, files, ext );

   SrString d;
   while ( dirs.size()>0 )
    { d = dirs[0];
      if ( d.last_char()!='/' ) d.append("/");
      //sr_out<<d<<srnl;
      dirs.remove ( 0 );
      fl_scan_dir ( d, dirs, files, ext );
    }
 }

//============================== End of File ==============================================
