
# include <FL/Fl.H>

# include <SR/sr_mat.h>
# include <SR/sr_matrix_window.h>

# include "sr_fl_ui.h"

//============================== SrMWin ==============================================

class SrMWin : public UiMatrixWin
 { public :
    SrMat mat;
    SrMatrixWindow* win;
   public :
    SrMWin ( SrMatrixWindow* mw );
    void update_ui ();
    void update_mat ();
    void update_max_min_step ();
    void mat_changed () { win->matrix_changed ( mat ); }
    virtual void event ( Event e );
 };

SrMWin::SrMWin ( SrMatrixWindow* mw )
 {
   win = mw;
   update_max_min_step();
   update_ui();
 }

void SrMWin::update_ui ()
 {
   int i;
   const double prec = 0.0000001;
   double d;
   for ( i=0; i<16; i++ )
    { d = (double)mat[i];
      if ( SR_BOUNDED(d,-prec,prec) ) d=0;
      val[i]->value ( d );
    }
 }

void SrMWin::update_mat ()
 {
   int i;
   for ( i=0; i<16; i++ ) mat[i] = (float) val[i]->value();
   win->matrix_changed ( mat ); 
 }

void SrMWin::update_max_min_step ()
 {
   int i;
   for ( i=0; i<16; i++ )
    { val[i]->minimum ( min_input->value() );
      val[i]->maximum ( max_input->value() );
      val[i]->step ( step_input->value() );
    }
 }

void SrMWin::event ( Event e )
 {
   SrMat r;
   float ang = SR_TORAD ( float(ang_value->value()) );
   switch ( e )
    { case evValChanged : update_mat(); break;
      case evOk :         win->ok_button(mat); break;
      case evCancel :     win->cancel_button(); break;
      case evIdentity :   mat.identity(); mat_changed(); update_ui(); break;
      case evTranspose :  mat.transpose(); mat_changed(); update_ui(); break;
      case evInverse :    mat.inverse(); mat_changed(); update_ui(); break;
      case evRotX :       r.rotx(ang); mat*=r; mat_changed(); update_ui(); break;
      case evRotY :       r.roty(ang); mat*=r; mat_changed(); update_ui(); break;
      case evRotZ :       r.rotz(ang); mat*=r; mat_changed(); update_ui(); break;
      case evMin:
      case evMax:
      case evStep:        update_max_min_step (); break;
    }
 }

//============================== SrMatrixWindow ==============================================

SrMatrixWindow::SrMatrixWindow ()
 {
   _ui = new SrMWin ( this );
   _ok_pressed = false;
 }

SrMatrixWindow::~SrMatrixWindow ()
 {
   delete _ui;
 }

void SrMatrixWindow::title ( const char* s )
 {
   if ( s ) _ui->window->label ( s );
 }

Fl_Window* SrMatrixWindow::window ()
 {
   return _ui->window;
 }

void SrMatrixWindow::mat ( const SrMat& m )
 {
   _ui->mat = m;
   _ui->update_ui ();
   matrix_changed ( _ui->mat );
 }

const SrMat& SrMatrixWindow::mat () const
 {
   return _ui->mat;
 }

void SrMatrixWindow::ok_button ( const SrMat& m )
 {
   _ui->window->hide();
   _ok_pressed = true;
 }

void SrMatrixWindow::cancel_button ()
 {
   _ui->window->hide();
   _ok_pressed = false;
 }

//============================== End of File ==============================================

