
# include <FL/Fl.H>

# include <SR/sr_var_table.h>
# include <SR/sr_var_table_window.h>

# include "sr_fl_ui.h"

//============================== SrVTWin ==============================================

class SrVTWin : public UiVarTableWin
 { public :
    SrVarTable* vt;
    SrVarTableWindow* win;
   public :
    SrVTWin ( SrVarTableWindow* w );
   ~SrVTWin () { vt->unref(); }
    void update_ui_all ();
    void update_ui ();
    void update_var ();
    virtual void event ( Event e );
 };

SrVTWin::SrVTWin ( SrVarTableWindow* w )
 {
   vt = new SrVarTable;
   vt->ref();
   win = w;
   update_ui();
 }

# define MAXL 80

void SrVTWin::update_ui ()
 {
   int i = browser->value()-1;
   if ( i<0 ) return;
   SrVar& v = vt->get(i);

   SrString s;
   SrOutput o(s);
   o << vt->get(i);
   if ( s.len()>MAXL ) { s[MAXL]=0; s<<"..."; }
   browser->text ( i+1, s );

   char c = v.type();

   type_output->value ( c=='b'? "bool" :
                        c=='i'? "int" :
                        c=='f'? "float" : "string" );

   name_output->value ( v.name() );

   value_text_input->hide();
   value_bool_choice->hide();
   value_num_input->hide();

   if ( v.size()>1 )
    { int val = (int)value_index->value();
      if ( val>=v.size() ) value_index->value(v.size()-1);
      value_index->activate();
      value_index->minimum(0);
      value_index->maximum(v.size()-1);
    }
   else
    { value_index->value(0);
      value_index->deactivate();
    }

   if ( c=='b' )
    { value_bool_choice->show();
      value_bool_choice->value ( v.getb() );
    }
   else if ( c=='i' )
    { value_num_input->show();
      value_num_input->value ( v.geti() );
    }
   else if ( c=='f' )
    { value_num_input->show();
      value_num_input->value ( v.getf() );
    }
   else if ( c=='s' )
    { value_text_input->show();
      value_text_input->value ( v.gets() );
    }
 }

void SrVTWin::update_ui_all ()
 {
   browser->clear ();
 
   int i;
   SrString s;
   SrOutput o(s);

   for ( i=0; i<vt->size(); i++ )
    { s.set(0);
      o << vt->get(i);
      if ( s.len()>MAXL ) { s[MAXL]=0; s<<"..."; }
      browser->add ( s );
    }

   browser->value(1);
   update_ui ();
 }

void SrVTWin::update_var ()
 {
   int i = browser->value()-1;
   if ( i<0 || i>=vt->size() ) return;

   SrVar& v = vt->get(i);
   char t = v.type();
   
   int ind = 0;
   if ( value_index->visible() )
    { ind = (int)value_index->value(); }

   if ( t=='b' )
    { v.set ( value_bool_choice->value()? true:false, ind );
    }
   else if ( t=='i' )
    { v.set ( (int)value_num_input->value(), ind );
    }
   else if ( t=='f' )
    { v.set ( (float)value_num_input->value(), ind );
    }
   else // s
    { v.set ( value_text_input->value(), ind );
    }
 }

void SrVTWin::event ( Event e )
 {
   switch ( e )
    { case evBrowser : update_ui(); break;
      case evModifyValue : 
           { int i = browser->value()-1;
             if ( i<0 || i>=vt->size() ) break;
             update_var();
             update_ui();
             win->value_modified(i);
           } break;
      case evClose : win->window_closed(); break;
    }
 }

//============================== SrVarTableWindow ==============================================

SrVarTableWindow::SrVarTableWindow ()
 {
   _ui = new SrVTWin ( this );
 }

SrVarTableWindow::~SrVarTableWindow ()
 {
   delete _ui;
 }

void SrVarTableWindow::title ( const char* s )
 {
   if ( s ) _ui->window->label ( s );
 }

Fl_Window* SrVarTableWindow::window ()
 {
   return _ui->window;
 }

void SrVarTableWindow::show ()
 {
   _ui->window->show();
 }

void SrVarTableWindow::use_var_table ( SrVarTable* vt )
 {
   if ( !vt ) vt = new SrVarTable;
   vt->ref();
   _ui->vt->unref();
   _ui->vt = vt;
   _ui->update_ui_all();
 }

void SrVarTableWindow::update ()
 {
   _ui->update_ui_all();
 }

void SrVarTableWindow::window_closed ()
 {
   _ui->window->hide();
 }

//============================== End of File ==============================================


