
# include <FL/fl_ask.H>
# include <FL/fl_file_chooser.H>
# include <FL/Fl_Color_Chooser.H>

# include "srm_main.h"
# include "srm_main_win.h"
# include "srm_material_tab.h"

# include <SR/sr_gl.h>
# include <SR/sr_fl.h>
# include <SR/sr_box.h>
# include <SR/sr_string.h>
# include <SR/sr_model.h>
# include <SR/sr_sphere.h>
# include <SR/sr_cylinder.h>
# include <SR/sr_sn_shape.h>
# include <SR/sr_sn_group.h>
# include <SR/sr_sn_manipulator.h>

//# define SR_USE_TRACE1 
# include <SR/sr_trace.h>

void SrmMatWin::matrix_changed ( const SrMat& m )
 {
   manip->initial_mat(m);
   viewer->render();
 }

void SrmMatWin::cancel_button ()
 {
   manip->initial_mat(restore_mat);
   viewer->render();
 }

SrmMainWin::SrmMainWin ()
 {
   _last_crease_angle_deg = 35.0f;
   _last_normals_prec_deg =  1.0f;
   _matwin = new SrmMatWin;
   _matwin->viewer = ui_viewer;
 }

SrmMainWin::~SrmMainWin ()
 {
   delete _matwin;
 }

void SrmMainWin::update_ui ()
 {
   int i = ui_browser->value()-1;

   bool model, normals, manip;
   ui_viewer->get_visibility ( i, model, normals, manip );
   ui_view_model_toggle->value ( model );
   ui_view_normals_toggle->value ( normals );
   ui_view_manip_toggle->value ( manip );

   ui_statistics_output->value ( "" );

   SrModel* m = cur_model();
   if ( !m )
    { ui_V->value ( 0 );
      ui_F->value ( 0 );
      ui_M->value ( 0 );
      ui_N->value ( 0 );
      ui_T->value ( 0 );
      ui_Ft->value ( 0 );
      ui_Fn->value ( 0 );
      ui_Fm->value ( 0 );
      ui_name_input->value ( "" );
      ui_culling_toggle->value ( 0 );
      return;
    }

   ui_V->value ( m->V.size() );
   ui_F->value ( m->F.size() );
   ui_M->value ( m->M.size() );
   ui_N->value ( m->N.size() );
   ui_T->value ( m->T.size() );
   ui_Ft->value ( m->Ft.size() );
   ui_Fn->value ( m->Fn.size() );
   ui_Fm->value ( m->Fm.size() );

   ui_name_input->value ( m->name );
   ui_culling_toggle->value ( m->culling );

   TheMaterialTab->set_smodel ( cur_smodel(), ui_browser->text(i+1) );

   if ( _matwin->window()->shown() ) // update window values
    { SrString title ( "Edit Transformation " ); title << i;
      _matwin->title ( title );
      _matwin->manip = ui_viewer->get_manipulator ( i );
      _matwin->mat ( _matwin->manip->mat() );
      _matwin->restore_mat = _matwin->manip->mat();
    }

   if ( ui_statistics_toggle->value() )
    { SrString s;
      SrOutput i(s); i.fmt_float("%.2f");
      SrBox box;
      SrArray<int> E;
      m->make_edges ( E );
      m->get_bounding_box ( box );
      int v = m->V.size();
      int e = E.size();
      int f = m->F.size();
      int fn = m->Fn.size();

      i << "min coords: " << box.a << srnl;
      i << "max coords: " << box.b << srnl;
      i << "mean vertex degree = " << m->count_mean_vertex_degree() << srnl;
      i << "V=" << v << "  E=" << e << "  F=" << f << "  Fn=" << fn << srnl;
      i << "V-E+F = " << (v-e+f) << " = h0-h1+h2\n";
      i << "h0 = shells\n";
      i << "h1 = 2*genus\n";
      i << "h2 = orientability (=h0)\n";
      ui_statistics_output->value ( s );
   }
 }

void SrmMainWin::add_model ( SrSnModel* sm, const char* filename )
 {
   SrString s, name;

   _lastdir = filename;
   _lastdir.remove_file_name();

   s = filename;
   s.get_file_name ( name );
   name.lower();
   s = "";

   int size = ui_browser->size();
   s << size << ": [" << name << "]";
   ui_viewer->add_model ( sm );
   ui_browser->add ( s );
   ui_browser->value ( size+1 );
   update_ui ();
 }

bool SrmMainWin::open ( const char *file )
 {
   SrString s, ext;

   if ( !file )
    { file = fl_file_chooser ( "Load Model", "*.srm|*.obj", 0 );
      if ( !file ) return false;
    }

   s = file;
   s.get_file_extension ( ext );
   ext.lower();

   SrSnModel* m = new SrSnModel;
   bool error = false;

   if ( ext=="srm" )
    { FILE *stream;
	  fopen_s(&stream,file,"rt");
	  SrInput in ( stream );
      if ( !in.valid() )
       { fl_alert ( "Could not open [%s]", file ); error=true; }
      else
       { if ( !m->shape().load(in) )
          { fl_alert ( "Error loading [%s]", file ); error=true; }
       }
    }
   else if ( ext=="obj" )
    { if ( !m->shape().import_obj(file) )
       { fl_alert ( "Error importing .obj file [%s]", file ); error=true; }
    }
   else
    { fl_alert ( "Unrecognized extension [%s]", (const char*)ext ); error=true; }

   if ( error )
    { delete m;
      return false;
    }
   else
    { add_model ( m, file );
      return true;
    }
 }

SrSnModel* SrmMainWin::cur_smodel () const
 {
   int i = ui_browser->value()-1;
   return ui_viewer->get_smodel ( i );
 }

SrModel* SrmMainWin::cur_model () const
 {
   int i = ui_browser->value()-1;
   return ui_viewer->get_model ( i );
 }

void SrmMainWin::cur_close ()
 {
   int i = ui_browser->value();
   if ( i<=0 ) return;
   bool ok = ui_viewer->remove_model ( i-1 );
   if ( !ok ) return;

   ui_browser->remove ( i );
   if ( i>ui_browser->size() ) i--;
   if ( i>0 ) ui_browser->value ( i );

   SrString s;
   for ( i=1; i<=ui_browser->size(); i++ )
    { const char* spt = ui_browser->text(i);
      while ( *spt>='0' && *spt<='9' ) spt++;
      s = "";
      s << (i-1) << spt;
      ui_browser->text ( i, s );
    }
 }

void SrmMainWin::open_numbered ( const char* first_file )
 {
   if ( !first_file )
     first_file = fl_file_chooser ( "Open First (nameNNNN.srm)", "*.srm", 0 );
   if ( !first_file ) return;

   SrString sn, s(first_file);
   int len = s.len();
   if ( len<9 ) return;
   sn=s;
   sn.substring(len-8,len-4);
   int n = sn.atoi();
   s[len-8]=0;
   FILE* fp;
   while ( true )
    { sn.setf ( "%s%04d.srm", (const char*)s, n );
      fopen_s ( &fp, sn, "r" );
      if ( !fp ) break; else fclose(fp);
      open(sn);
      n++;
    }
 }

void SrmMainWin::open_dir ( const char* folder )
 {
   SrString dir;
   SrStringArray dirs;
   SrStringArray files;
   SrStringArray ext;
   
   if ( !folder )
    folder = fl_dir_chooser ( "Select Folder", 0, 0 );
   if ( !folder ) return;
   dir = folder;
   
      
   if ( dir=="" ) return;
   ext.push ( "srm" );
   if ( dir.last_char()!='/' ) dir.append ("/");
   fl_scan_dir ( dir, dirs, files, ext );

   int i;
   for ( i=0; i<files.size(); i++ )
    { open ( files[i] );
    }
 }

void SrmMainWin::add_primitive ( SrmMainEvent type, const char* coords )
 {
   const char* action;
   const char* prim;
   const char* desc;
   SrString s;
   bool ok = true;

   if ( coords ) s=coords;
   else
   { switch ( type )
     { case evMakeBox :
        { action = "Make Box";
          prim = "box";
          desc = "Enter min/max coordinates:\n(6 space separated numbers)";
        } break;
       case evMakeSphere :
        { action = "Make Sphere";
          prim = "sphere";
          desc = "Enter center, radius, and resolution:\n(5 space separated numbers)";
        } break;
       case evMakeCylinder :
        { action = "Make Cylinder";
          prim = "cylinder";
          desc = "Enter the two endpoints, radius, and resolution:\n(8 space separated numbers)";
        } break;
     };

    ok = fl_string_input ( action, desc, s );
    if ( !ok ) return;
   }

   SrSnModel* newsm = new SrSnModel;

   SrInput in(s);
   float res=1.0f;
   
   switch ( type )
    { case evMakeBox :
       { SrBox b; in >> b;
         if ( b.a==b.b ) ok=false; else newsm->shape().make_box ( b );
       } break;
      case evMakeSphere :
       { SrSphere s; in >> s >> res;
         if ( s.radius<=0 ) ok=false; else newsm->shape().make_sphere ( s, res );
         fl_alert ( "Make Sphere Not Implemented!" );
       } break;
      case evMakeCylinder :
       { SrCylinder c; in >> c >> res;
         if ( c.radius<=0 ) ok=false; else newsm->shape().make_cylinder ( c, res );
       } break;
    };

   if ( !ok || in.last_error()!=SrInput::NoError )
    { fl_alert ( "Invalid Input!" );
      delete newsm;
    }
   else
    { newsm->shape().M.push();
      SrMaterial mtl;
      newsm->shape().set_one_material ( mtl );
      add_model ( newsm, prim );
    }
 }

void SrmMainWin::event ( SrmMainEvent e )
 {
   switch ( e )
    { case evFileOpen : open(); break;

      case evFileSave :
       { SrModel* m = cur_model();
         if ( !m ) break;
         const char* cpt = ui_browser->text(ui_browser->value());
         while ( *cpt!='[' ) cpt++; cpt++;
         SrString fname;
         fname << _lastdir << cpt;
         fname.last_char(0);
         char *s = fl_file_chooser ( "Save SrModel", "*.srm", fname );
         if ( !s ) break;
		 FILE *stream;
		 fopen_s(&stream,s,"wt");
         SrOutput out ( stream );
         if ( !out.valid() ) fl_alert("Could not write file!");
          else m->save ( out );
         break;
       }

      case evFileExport :
       { SrModel* m = cur_model();
         if ( !m ) break;
         char *s = fl_file_chooser ( "Export Model", "*.iv", 0 );
         if ( !s ) break;
         SrString file(s), ext;
         file.get_file_extension ( ext );
         if ( ext=="iv" )
          { if ( !m->export_iv(s) )
             fl_alert("Error exporting [%s]", s);
          }
         else
          fl_alert("Error: extension not supported!");
         break;
       }

      case evFileClose :
       { cur_close ();
         update_ui ();
         ui_viewer->render();
         break;
       }

      case evFileOpenNumbered :
       { open_numbered ();
       } break;

      case evFileOpenDir :
       { open_dir ();
       } break;

      case evFileCloseAll :
       { while ( ui_browser->size()>0 ) cur_close();
         update_ui ();
         ui_viewer->render();
       } break;

      case evEditInvertFaces :
       { SrModel* m = cur_model();
         if ( !m ) break;
         m->invert_faces();
         update_ui ();
         ui_viewer->render();
         break;
       }

      case evEditInvertNormals :
       { SrModel* m = cur_model();
         if ( !m ) break;
         m->invert_normals();
         update_ui ();
         ui_viewer->reset_normals ( ui_browser->value()-1 );
         ui_viewer->render();
         break;
       }

      case evEditFlatNormals :
       { SrModel* m = cur_model();
         if ( !m ) break;
         m->flat();
         ui_viewer->reset_normals ( ui_browser->value()-1 );
         ui_viewer->render();
         update_ui ();
         break;
       }

      case evEditSmoothNormals :
       { SrModel* m = cur_model();
         if ( !m ) break;
         bool ok = fl_value_input ( "Smooth Normals", 
                                    "Crease Angle in Degrees (-1 for overall smooth):",
                                     _last_crease_angle_deg );
         if ( !ok ) break;
         if ( _last_crease_angle_deg==-1.0f ) 
          m->smooth ( _last_crease_angle_deg );
         else
          m->smooth ( SR_TORAD(_last_crease_angle_deg) );
         update_ui ();
         ui_viewer->reset_normals ( ui_browser->value()-1 );
         ui_viewer->render();
         break;
       }

      case evEditNormalize :
       { SrModel* m = cur_model();
         if ( !m ) break;
         float mc = 1;
         bool ok = fl_value_input ( "Normalize", 
                                    "Enter max coord:",
                                     mc );
         if ( !ok && mc<=0 ) break;
         m->normalize ( mc );
         update_ui ();
         ui_viewer->reset_normals ( ui_browser->value()-1 );
         ui_viewer->render();
         break;
       }

      case evEditValidate :
       { SrModel* m = cur_model();
         if ( !m ) break;
         bool ok = fl_value_input ( "Validation",
                                    "Normals Angular Precision (deg):",
                                     _last_normals_prec_deg );
         if ( !ok ) break;
         m->validate ();
         m->remove_redundant_materials ();
         m->remove_redundant_normals ( SR_TORAD(_last_normals_prec_deg) );
         m->compress ();
         update_ui ();
         ui_viewer->reset_normals ( ui_browser->value()-1 );
         ui_viewer->render();
         break;
       }

      case evCullingToggle :
       { SrModel* m = cur_model();
         if ( !m ) break;
         m->culling = ui_culling_toggle->value()? true:false;
         ui_viewer->render();
         break;
       }

      case evNameInput :
       { SrModel* m = cur_model();
         if ( !m ) break;
         m->name = ui_name_input->value();
         break;
       }

      case evViewChange :
       ui_viewer->set_visibility ( ui_browser->value()-1,
                                   ui_view_model_toggle->value()?1:0,
                                   ui_view_normals_toggle->value()?1:0,
                                   ui_view_manip_toggle->value()?1:0 );
       ui_viewer->render();
       break;

      case evModelChange :
       update_ui ();
       Fl::check();
       break;

      case evStatisticsToggle :
       update_ui ();
       Fl::check();
       break;

      case evJoinModels :
       { SrString s;
         bool ok = fl_string_input ( "Join Models", "Enter ids of the models to join:", s );
         if ( !ok ) break;
         int i;
         SrInput in(s);
         SrArray<int> ids;
         while ( true )
          { if ( in.get_token()==SrInput::Integer ) ids.push()=atoi(in.last_token());
            if ( in.finished() ) break;
          }
         in.close();
         s = "";
         if ( ids.size()==0 ) break;
         SrSnModel* sm;
         SrSnModel* newsm = new SrSnModel;
         for ( i=0; i<ids.size(); i++ )
          { sm = ui_viewer->get_smodel(ids[i]);
            if ( sm )
             { newsm->shape().add_model ( sm->const_shape() );
               s << ids[i] << '+';
             }
          }
         if ( s.len()>0 ) s.last_char(0);
         if ( newsm->shape().empty() ) delete newsm;
          else add_model ( newsm, s );
       }
      break;

      case evMakeBox :
      case evMakeSphere :
      case evMakeCylinder :
       { add_primitive ( e );
       } break;

      case evInfoOpenGL :
       { SrString s;
         SrOutput o(s);
         glPrintInfo ( o );
         fl_text_output ( "OpenGL Properties", s );
         break;
       }

      case evManipApplyToModel :
       { SrModel* model = cur_model();
         SrSnManipulator* manip = ui_viewer->get_manipulator ( ui_browser->value()-1 );
         if ( !model || !manip ) break;
         model->apply_transformation ( manip->mat() );
         manip->initial_mat ( SrMat::id );
         if ( _matwin->window()->shown() )
          { _matwin->restore_mat=SrMat::id; _matwin->mat(SrMat::id); }
         ui_viewer->render();
         break;
       }

      case evManipEditTransform :
       { if ( _matwin->window()->shown() ) break;
         int i = ui_browser->value()-1;
         SrSnManipulator* manip = ui_viewer->get_manipulator ( i );
         if ( !manip ) break;
         SrString title ( "Edit Transformation " ); title << i;
         _matwin->manip = manip;
         _matwin->title ( title );
         _matwin->mat ( manip->mat() );
         _matwin->restore_mat = manip->mat();
         _matwin->window()->show();
         break;
       }

      case evViewFirst :
       { int i;
         for ( i=1; i<ui_viewer->models(); i++ ) ui_viewer->set_visibility(i,false,false,false);
         ui_viewer->set_visibility(0,true,false,false);
         ui_browser->value(1);
         update_ui ();
         ui_viewer->render();
       } break;
      
      case evViewPrior :
       { int i = ui_browser->value()-1;
         if ( i<=0 ) break;
         ui_viewer->set_visibility(i,false,false,false);
         ui_viewer->set_visibility(i-1,true,false,false);
         ui_browser->value(i);
         update_ui ();
         ui_viewer->render();
       } break;
       
      case evViewNext :
       { int i = ui_browser->value()-1;
         if ( i<0 || i>=ui_viewer->models() ) break;
         ui_viewer->set_visibility(i,false,false,false);
         ui_viewer->set_visibility(i+1,true,false,false);
         ui_browser->value(i+2);
         update_ui ();
         ui_viewer->render();
       } break;

      case evTabGroup :
       { if ( ui_tab_group->value()==ui_material_group )
          TheMaterialTab->show();
       } break;
    }
 }

void SrmMainWin::event ( SrmMtlTabEvent e )
 {
   TheMaterialTab->event ( e );
 }
