
# include <FL/Fl_Color_Chooser.H>

# include <SR/sr_fl.h>
# include <SR/sr_model.h>
# include <SR/sr_string_array.h>
# include <SR/sr_sn_shape.h>

# include "srm_main_win.h"
# include "srm_material_tab.h"
# include "srm_main.h"

SrmMaterialTab::SrmMaterialTab ( SrmMainWin* mw )
 {
   _changed = true;
   _model = 0;
   _smodel = 0;
   _w = mw;
 }

void SrmMaterialTab::set_smodel ( SrSnModel* sm, const char* text )
 {
   SrSnModel* old_smodel = _smodel;
   _smodel = sm;
   _model = &sm->shape();

   _w->ui_mtlmodel_output->value ( text );
   
   if ( old_smodel!=sm ) set_changed();
 }

void SrmMaterialTab::set_changed ()
 {
   _changed = true;
   if ( _w->ui_window->shown() ) show ();
 }

void SrmMaterialTab::show ()
 {
   if ( !_model )
    { _w->ui_mtlbrowser->clear(); _changed=true; return; }

   if ( _changed )
    { SrString s;
      SrStringArray& names = _model->mtlnames;
      if ( names.size()==0 ) names.size(_model->M.size() );
      _w->ui_mtlbrowser->clear();
      int i=0;
      for ( i=0; i<_model->M.size(); i++ )
       { s = "";
         s << i << ": [" << names[i] << "]";
         _w->ui_mtlbrowser->add(s);
       }
      _w->ui_mtlbrowser->value(1);
      update_dialog (0);
      _changed = false;
    }

   _w->ui_window->show ();
   _w->ui_window->redraw ();
 }

//  if ( !fl_color_chooser("Select Color:",pt[0],pt[1],pt[2]) ) return;

static void setcolor ( Fl_Widget *w, int index, SrColor c,
                       Fl_Value_Input *r, Fl_Value_Input *g, Fl_Value_Input *b, Fl_Value_Input *a )
 {
   Fl_Color x = (Fl_Color)index;
   Fl::set_color ( x, c.r, c.g, c.b );
   w->color ( x, x );
   w->redraw ();
   r->value ( c.r );
   g->value ( c.g );
   b->value ( c.b );
   a->value ( c.a );
 }

void SrmMaterialTab::update_dialog ( Fl_Button *but )
 {
   int p = _w->ui_mtlbrowser->value()-1;
   if ( p<0 ) 
    { _w->ui_mtlbrowser->value(1); p=0; }

   SrMaterial m;
   if ( _model->M.size() ) m = _model->M[p];
   if ( !but || but==_w->ui_ambient ) setcolor ( _w->ui_ambient,  200, m.ambient,   _w->ui_ar, _w->ui_ag, _w->ui_ab, _w->ui_aa );
   if ( !but || but==_w->ui_diffuse ) setcolor ( _w->ui_diffuse,  201, m.diffuse,   _w->ui_dr, _w->ui_dg, _w->ui_db, _w->ui_da );
   if ( !but || but==_w->ui_specular ) setcolor ( _w->ui_specular, 202, m.specular, _w->ui_sr, _w->ui_sg, _w->ui_sb, _w->ui_sa );
   if ( !but || but==_w->ui_emission ) setcolor ( _w->ui_emission, 203, m.emission, _w->ui_er, _w->ui_eg, _w->ui_eb, _w->ui_ea );
   if ( !but ) _w->ui_shininess->value ( m.shininess );
 }

void SrmMaterialTab::update_model ( Fl_Button *but )
 {
   int p = _w->ui_mtlbrowser->value()-1;
   if ( p<0 || !_model ) return;
   SrMaterial &m = _model->M[p];

   if ( but==_w->ui_ambient )
    { SrColor &c = m.ambient;
      c.set ( (srbyte)_w->ui_ar->value(), (srbyte)_w->ui_ag->value(), (srbyte)_w->ui_ab->value(), (srbyte)_w->ui_aa->value() );
      Fl::set_color ( (Fl_Color)200, c.r, c.g, c.b );
    }
   else if ( but==_w->ui_diffuse  )
    { SrColor &c = m.diffuse;
      c.set ( (srbyte)_w->ui_dr->value(), (srbyte)_w->ui_dg->value(), (srbyte)_w->ui_db->value(), (srbyte)_w->ui_da->value() );
      Fl::set_color ( (Fl_Color)201, c.r, c.g, c.b );
    }
   else if ( but==_w->ui_specular  )
    { SrColor &c = m.specular;
      c.set ( (srbyte)_w->ui_sr->value(), (srbyte)_w->ui_sg->value(), (srbyte)_w->ui_sb->value(), (srbyte)_w->ui_sa->value() );
      Fl::set_color ( (Fl_Color)202, c.r, c.g, c.b );
    }
   else if ( but==_w->ui_emission  )
    { SrColor &c = m.emission;
      c.set ( (srbyte)_w->ui_er->value(), (srbyte)_w->ui_eg->value(), (srbyte)_w->ui_eb->value(), (srbyte)_w->ui_ea->value() );
      Fl::set_color ( (Fl_Color)203, c.r, c.g, c.b );
    }
   else
    { m.shininess = (srbyte)_w->ui_shininess->value();
    }

   if ( but ) but->redraw ();

   if ( _w->ui_update_but->value() )
    { _smodel->changed (true);
      _w->ui_viewer->render();
    }
 }

void SrmMaterialTab::new_color ( Fl_Button *but )
 {
   if ( !_model ) return;

   int p = _w->ui_mtlbrowser->value()-1;
   if ( p<0 ) return;
   SrMaterial &m = _model->M[p];

   SrColor *c;
   if ( but==_w->ui_ambient  ) c=&m.ambient;
   else if ( but==_w->ui_diffuse  ) c=&m.diffuse;
   else if ( but==_w->ui_specular ) c=&m.specular;
   else if ( but==_w->ui_emission ) c=&m.emission;

   if ( !fl_color_chooser("New Color:",*c) ) return;

   update_dialog ( but );
   update_model ( but );
 }

void SrmMaterialTab::event ( SrmMtlTabEvent e )
 {
   switch ( e )
    { case evBrowser : update_dialog(0); break;
       
      case evAmbient : new_color(_w->ui_ambient); break;
      case evDiffuse : new_color(_w->ui_diffuse); break;
      case evSpecular : new_color(_w->ui_specular); break;
      case evEmission : new_color(_w->ui_emission); break;

      case evAmbientValue : update_model ( _w->ui_ambient ); break;
      case evDiffuseValue : update_model ( _w->ui_diffuse ); break;
      case evSpecularValue : update_model ( _w->ui_specular ); break;
      case evEmissionValue : update_model ( _w->ui_emission ); break;

      case evClearAll :
           _model->clear_materials();
           _changed = true;
           show ();
           break;

      case evSetOne :
           { SrMaterial mtl;
             _model->set_one_material ( mtl );
             _changed = true;
             show ();
           } break;

      case evAddOne :
           { SrString name ("material");
             if ( !fl_string_input("New Material Name", "Enter Name:", name ) ) break;
             _model->mtlnames.push ( name );
             _model->M.push().init();
             _changed = true;
             show ();
           } break;

      case evShininessValue :
      case evUpdate :
       { update_model(0); 
         break;
       }
    }
 }
