
# ifndef SRM_VIEWER_H
# define SRM_VIEWER_H

# include <SR/sr_viewer.h>
# include <SR/sr_sn_group.h>
# include <SR/sr_sn_shape.h>

class SrSnManipulator;

// maintained scene graph :
//
//                 root
//       |          |        |
//  manipulator    ...   manipulator
//       |                   |
//     group               group
//   |       |            |        |   
// model  normals       model   normals

class SrmViewer : public SrViewer
 { private :
    SrSnGroup* _root;
    int _lastfac, _lastmtl;
   public :
    SrmViewer ( int x, int y, int w, int h );
    void add_model ( SrSnModel* model );
    bool remove_model ( int i );

    SrSnModel* get_smodel ( int i );
    SrModel* get_model ( int i );
    SrSnManipulator* get_manipulator ( int i );

    int models () { return _root->size(); }

    void reset_normals ( int i );
    void make_normals ( int i );

    void set_visibility ( int i, bool model, bool normals, bool manip );
    void get_visibility ( int i, bool& model, bool& normals, bool& manip );

   public :
    virtual int handle_keyboard ( const SrEvent& e );
    virtual int handle_scene_event ( const SrEvent& e );
 };

# endif // SRM_VIEWER_H
