
# include <string.h>

# include <SR/sr_timer.h>
# include <SR/sr_array.h>
# include <SR/sr_list.h>
# include <SR/sr_input.h>
# include <SR/sr_output.h>

class Buf : public SrListNode
 { public :
    char st[80];
    Buf ( char *s ) { strcpy_s(st,80,s); }
    Buf () { st[0]=0; }
    void operator = ( const Buf &b ) { strcpy_s(st,80,b.st); }
    friend SrOutput& operator<< ( SrOutput& o, const Buf &b ) { return o<<b.st; }
    friend SrInput& operator>> ( SrInput& i, Buf& b ) { return i>>(b.st); }
    friend int sr_compare  ( const Buf *b1, const Buf *b2 ) { return sr_compare(b1->st,b2->st); }
 };

# define SIZE 1000

static void run ()
 {
   SrArray<Buf> a;
   SrList<Buf>  l;

   sr_out << "Time required for different structure manipulations\n";
   sr_out << "Manipulating " << SIZE << " elements\n\n";

   int i;
   double t1, t2;
   SrTimer t;

   t.start();
   for ( i=0; i<SIZE; i++ ) a.push("XXX");
   t.stop(); t1=t.dt();
   t.start();
   for ( i=0; i<SIZE; i++ ) l.push_back ( new Buf("XXX") );
   t.stop(); t2=t.dt();
   sr_out << "1. Insertion: Array: "<<t1<<"  List: "<<t2<< '\n';

   t.start();
   a.sort (sr_compare);
   t.stop(); t1=t.dt();
   t.start();
   l.sort ();
   t.stop(); t2=t.dt();
   sr_out << "2. Sort:      Array: "<<t1<<"  List: "<<t2<< '\n';

   t.start();
   for ( i=0; i<SIZE; i++ ) a.remove(0);
   t.stop(); t1=t.dt();
   t.start();
   for ( i=0; i<SIZE; i++ ) delete l.pop_back();
   t.stop(); t2=t.dt();
   sr_out << "3. Deletion:  Array: "<<t1<<"  List: "<<t2<< '\n';
 }

int arraylist_main ( int, char ** )
 {
   run ();
   return 0;
 }

/* SIZE==5000, Impact, with a.remove(0) :
1. Insertion: Array: 0.010340 List: 0.016021 ! even with array reallocs
2. Sort: Array: 0.083514 List: 5.082449
3. Deletion: Array: 2.794826 List: 0.008282
*/

/* SIZE==5000, Impact, with a.remove(i<2000? 0:1900) :
1. Insertion: Array: 0.009994 List: 0.016544 ! even with array reallocs
2. Sort: Array: 0.081935 List: 4.956194
3. Deletion: Array: 1.957149 List: 0.007001
*/
