
# include <stdio.h>
# include <stdlib.h>

# include <SR/sr_quat.h>
# include <SR/sr_mat.h>
# include <SR/sr_euler.h>
# include <SR/sr_random.h>
# include <SR/sr_timer.h>

const float EPS = 1.0E-5f;

bool test_euler ( int order, float rx, float ry, float rz )
 {
   SrMat m, eulerm;
   SrMat mx; mx.rotx(rx);
   SrMat my; my.roty(ry);
   SrMat mz; mz.rotz(rz);

   switch ( order )
    { case 123: m = mx*my*mz; break;
      case 132: m = mx*mz*my; break;
      case 213: m = my*mx*mz; break;
      case 231: m = my*mz*mx; break;
      case 312: m = mz*mx*my; break;
      case 321: m = mz*my*mx; break;
    }
   sr_euler_mat ( order, eulerm, rx, ry, rz );

   if ( !m.isnext(eulerm,EPS) )
    { sr_out<<"Mat Error "<<order<<":\n"; 
      m.round(EPS); eulerm.round(EPS);
      sr_out<<m<<srnl<<eulerm;
      sr_out<<srnl;
      return false;
    }

   float ex, ey, ez;
   sr_euler_angles ( order, m, ex, ey, ez );
   sr_euler_mat ( order, eulerm, ex, ey, ez );

   if ( !m.isnext(eulerm,EPS) )
    { sr_out<<"Angles Error "<<order<<":\n"; 
      sr_out<<SR_TODEG(rx)<<srspc<<SR_TODEG(ry)<<srspc<<SR_TODEG(rz)<<srnl;
      sr_out<<SR_TODEG(ex)<<srspc<<SR_TODEG(ey)<<srspc<<SR_TODEG(ez)<<srnl;
      sr_out<<srnl;
      m.round(EPS); eulerm.round(EPS);
      sr_out<<m<<srnl<<eulerm;
      sr_out<<srnl;
      return false;
    }
    
   return true;
 }

float ang ()
 {
   static SrRandom rang (-360,360);
   static SrRandom rf (0.0f,1.0f);
   static SrRandom ri (-4,4);
 
   if ( rf.getf()<0.75f ) // true random angle
    {
      return SR_TORAD(rang.geti());
    }
   else // special cases
    {
      return float(ri.geti())*srpidiv2;
    }
 }
 
int euler_main ( int, char ** )
 {
   int i;
   float rx, ry, rz;

   SrUtcTimer t;
   double t1 = t.time();
   int max = 100000;
   int error;
 
   for ( i=0; i<max; i++ )
    {
      rx = ang();
      ry = ang();
      rz = ang();
      error=0;
      if ( !test_euler ( 123, rx, ry, rz ) ) error++;
      if ( !test_euler ( 132, rx, ry, rz ) ) error++;
      if ( !test_euler ( 213, rx, ry, rz ) ) error++;
      if ( !test_euler ( 231, rx, ry, rz ) ) error++;
      if ( !test_euler ( 312, rx, ry, rz ) ) error++;
      if ( !test_euler ( 321, rx, ry, rz ) ) error++;

      if ( error>0 ) break;
      if ( i%5000==0 ) sr_out<<(max-i)<<"...\n";
      
    }

   double t2 = t.time();

   sr_out<<"Performed: "<<i<<" tests\n";
   sr_out<<"Total time: "<<(t2-t1)<<" secs\n";

   return 0;
 }
