

// uncomment this if you have glut installed in your system:
// # define GLUT_INSTALLED

# ifdef GLUT_INSTALLED

# include "srtest_glut.h"
# include <glut.h>

//============================= static data =================================

static SrGlutViewer* _cur_viewer=0;

static void glut_reshape ( int width, int height )
 { _cur_viewer->reshape ( width, height ); }

static void glut_keyboard ( unsigned char key, int x, int y )
 { _cur_viewer->keyboard ( key, x, y ); }

static void glut_mouse ( int button, int state, int x, int y )
 { _cur_viewer->mouse ( button, state, x, y ); }

static void glut_motion ( int x, int y )
 { _cur_viewer->motion ( x, y ); }

static void glut_display () 
 { _cur_viewer->display(); }

//============================= SrGlutViewer =================================

SrGlutViewer::SrGlutViewer ( const char* title )
 {
   _win_title = title;
   _win_dx = _win_dy = 100;
   _root = 0;
   mousex = mousey = lmousex = lmousey = 0;
 }

void SrGlutViewer::show ()
 {
   if ( !_cur_viewer ) // initialization
    { glutInitDisplayMode ( GLUT_RGBA | GLUT_DOUBLE );
    }

   _cur_viewer = this;
   glutCreateWindow ( _win_title );
   glutDisplayFunc  ( glut_display );
   glutKeyboardFunc ( glut_keyboard );
   glutMouseFunc    ( glut_mouse );
   glutReshapeFunc  ( glut_reshape );
   glutMotionFunc   ( glut_motion );

   glutPostRedisplay ();
 }

void SrGlutViewer::run ()
 {
   glutMainLoop();
 }

void SrGlutViewer::root ( SrSn *r )
 { 
   if ( _root ) _root->unref();
   if ( r ) r->ref();
   _root=r; 
 }

void SrGlutViewer::view_all ()
 {
   _cam.aspect = (float)_win_dx/(float)_win_dy;
   _ba.apply ( _root );
   _cam.view_all ( _ba.get(), SR_TORAD(60) );
   _tb.init ();
 }

void SrGlutViewer::reshape ( int width, int height )
 {
   sr_out<<"RESHAPE"<<srnl;
   _win_dx = width;
   _win_dy = height;
   glViewport ( 0, 0, _win_dx, _win_dy );
   glEnable ( GL_LIGHT0 ); 
   glEnable ( GL_LIGHTING );
   glEnable ( GL_DEPTH_TEST );
   glEnable ( GL_BLEND ); // for transparency
   glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

   glCullFace ( GL_BACK );
   glDepthFunc ( GL_LEQUAL );

   glFrontFace ( GL_CCW );
   glEnable ( GL_LINE_SMOOTH );
   glHint ( GL_LINE_SMOOTH_HINT, GL_NICEST );

   glEnable ( GL_POINT_SMOOTH );
   glPointSize ( 2.0 );

   glShadeModel ( GL_SMOOTH );
 }

void SrGlutViewer::keyboard ( unsigned char key, int x, int y )
 {
   switch ( key )
    { case 27  : exit(0); // ESC
      case 32  : view_all(); break;
      case '1' : _ra.override_render_mode ( _root, srRenderModeSmooth ); break;
      case '2' : _ra.override_render_mode ( _root, srRenderModeFlat ); break;
      case '3' : _ra.override_render_mode ( _root, srRenderModeLines ); break;
      case '4' : _ra.override_render_mode ( _root, srRenderModePoints ); break;
    }
   glutPostRedisplay ();
 }

void SrGlutViewer::mouse ( int button, int state, int x, int y )
 {
   if ( state==GLUT_DOWN )
    { float w = (float)glutGet(GLUT_WINDOW_WIDTH);
      float h = (float)glutGet(GLUT_WINDOW_HEIGHT);
      // get normalized coords in [-1,-1]x[1,1]
      mousex = lmousex =  (2*float(x)/w -1);
      mousey = lmousey = -(2*float(y)/h -1);
      sr_out<< mousex <<srspc<< mousey<<srnl;
    }
 }

void SrGlutViewer::motion ( int x, int y )
 {
   float w = (float)glutGet(GLUT_WINDOW_WIDTH);
   float h = (float)glutGet(GLUT_WINDOW_HEIGHT);
   // get normalized coords in [-1,-1]x[1,1]
   mousex =  (2*float(x)/w -1);
   mousey = -(2*float(y)/h -1);

   _tb.set_spin_from_mouse_motion ( lmousex, lmousey, mousex, mousey );
   SrQuat drot =_tb.last_spin.inverse();
   SrQuat rotation = _tb.rotation;
   _cam *= rotation.inverse() * drot * rotation;

   lmousex = mousex;
   lmousey = mousey;

   glutPostRedisplay ();
 }

void SrGlutViewer::display ()
 {
   if ( !_root ) return;
   glClearColor ( 1, 1, 1, 0 );
   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   glMatrixMode ( GL_PROJECTION );
   glLoadIdentity ();
   glMultMatrix ( _cam.get_perspective_mat(_mat) ); // set fovy/near/far/aspect

   glMatrixMode ( GL_MODELVIEW );
   glLoadIdentity ();
   glMultMatrix ( _cam.get_view_mat(_mat) ); // set eye/center/up

   _light.position = _cam.eye;               // set light
   _light.directional = false;
   glLight ( 0, _light );

   _ra.apply ( _root );                      // draw scene

   glutSwapBuffers ();
 }

//================================= Main =================================

# include <SR/sr_model.h>
# include <SR/sr_lines.h>
# include <SR/sr_sn_shape.h>
# include <SR/sr_sn_group.h>

int glut_main ( int argc, char **argv )
 {
   char *file = "../data/knot.srm";
   if ( argc>=2 ) file = argv[1];

   SrInput in ( fopen(file,"rt") );
   SrSnModel* smodel = new SrSnModel;
   SrModel& model = smodel->shape();
   if ( !in.valid() || !model.load(in) ) sr_out.fatal_error("Could not read model!\n");

   SrSnLines* sl = new SrSnLines;
   SrLines& l = sl->shape();
   l.push_axis ( SrPnt::null/*origin*/, 1/*length*/, 3/*dimension*/, "xyz"/*letters*/ );

   SrSnBox* sb = new SrSnBox;
   sb->render_mode ( srRenderModeLines );
   sb->can_override_render_mode(false);
   sb->color ( SrColor::gray );
   model.get_bounding_box ( sb->shape() );

   SrSnGroup* sg = new SrSnGroup;
   sg->add ( smodel );
   sg->add ( sb );
   sg->add ( sl );

   SrGlutViewer *viewer = new SrGlutViewer ( "SrGlutViewer Demo" );
   viewer->root ( sg );
   viewer->view_all ();
   viewer->show ();
   viewer->run ();

   return 0;
 }

# endif // GLUT_INSTALLED
