
# include <SR/sr_hash_table.h>

void print ( SrHashTableBase& T )
 {
   int i, j;
   const char* st;

   sr_out <<"\nHash Table:\n";

   for ( i=0; i<T.hash_size(); i++ )
    {
      sr_out << i << ":[";
      st = T.key(i);
      if ( !st ) st = "";
      sr_out << st << "]";
      j = i;
      while ( T.collidingid(j)!=-1 )
       { j = T.collidingid(j);
         sr_out << ", " << j << ":[" << T.key(j) << "]";
       }
      sr_out << srnl;
    }

   sr_out << "HashSize:"<<T.hash_size()
          << " Elements:"<<T.elements()
          << " LongestEntry:"<< T.longest_entry()
          << " Collisions:"<<T.collisions()<<"\n\n";
 }

struct MyData 
 { int x; 
   MyData(int i):x(i) {}
 };

int hash_table_main ( int, char ** )
 {
   SrHashTableBase TB(40);
   
   TB.insert ( "", 0 );
   TB.insert ( "atest", 0 );
   TB.insert ( "btest", 0 );
   TB.insert ( "ctest", 0 );
   TB.insert ( "abc", 0 );
   TB.insert ( "go now", 0 );
   TB.insert ( "lshoulder", 0 );
   TB.insert ( "lelbow", 0 );
   TB.insert ( "LWrist", 0 );
   TB.insert ( "lwrist", 0 );
   TB.insert ( "rwrist", 0 );
   TB.insert ( "rshoulder", 0 );
   TB.insert ( "relbow", 0 );
   TB.insert ( "l_shoulder", 0 );
   TB.insert ( "l_elbow", 0 );
   TB.insert ( "l_wrist", 0 );
   TB.insert ( "r_wrist", 0 );
   TB.insert ( "r_shoulder", 0 );
   TB.insert ( "r_elbow", 0 );

   print ( TB );

   TB.rehash ( 20 );
   print ( TB );
   
   sr_out<<"index of [relbow]: "<<TB.lookup_index("relbow")<<srnl;
   sr_out<<"index of [relbowX]: "<<TB.lookup_index("relbowX")<<srnl;
   
   TB.remove("ctest");
   TB.remove("abc");
   TB.remove("rshoulder");
   TB.remove("l_shoulder");
   sr_out<<"\nremoved [ctest,abc,rshoulder,l_shoulder]: "<<srnl;
   print ( TB );

   sr_out<<"\nInserting back [ctest,abc,rshoulder,l_shoulder]: "<<srnl;
   TB.insert ( "ctest", 0 );
   TB.insert ( "abc", 0 );
   TB.insert ( "rshoulder", 0 );
   TB.insert ( "l_shoulder", 0 );
   print ( TB );

   sr_out<<"\nTemplate version:\n\n";
   
   SrHashTable<MyData*> T(5);
   T.insertstat ( "d1", new MyData(1) );
   T.insertstat ( "d2", new MyData(2) );
   T.insertstat ( "d3", new MyData(3) );

   if ( !T.insertstat("d3",0) ) sr_out<<"duplication detected!\n";
   
   print ( T );
   sr_out << T.lookup("d1")->x << srnl;
   sr_out << T.lookup("d2")->x << srnl;
   sr_out << T.lookup("d3")->x << srnl;
   
   T.delete_data();
   
   return 0;
 }

