
# include <stdio.h>
# include <stdlib.h>

# include <SR/sr_matn.h>
# include <SR/sr_output.h>
# include <SR/sr_output_window.h>

void test_lu ( int n )
 {
   SrMatn a(n,n), l, u, lu;

   a.set_random ( 0, 99.0 );

   ludcmp ( a, l, u );
   sr_out << "a :\n" << a;
   sr_out << "l :\n" << l;
   sr_out << "u :\n" << u;
  
   lu.mult ( l, u );

   sr_out << "error : " << dist(a,lu) << srnl;
 }

void test_lusolve ( int n )
 {
   SrMatn a(n,n);
   SrMatn b(n,1);

   a.set_random ( 0, 99.0 );
   b.set_random ( 0, 99.0 );

   sr_out << "a :\n" << a;
   sr_out << "b :\n" << b;
  
   SrMatn x(b);
   lusolve ( a, b, x );

   sr_out << "x :\n" << x;

   SrMatn b2;
   b2.mult ( a, x );
   sr_out << "b2 :\n" << b2;

   sr_out << "error : " << dist(b2,b) << srnl;
 }

void test_gauss ( int n )
 {
   SrMatn a(n,n);
   SrMatn b(n,1);

   a.set_random ( 0, 99.0 );
   b.set_random ( 0, 99.0 );

   sr_out << "a :\n" << a;
   sr_out << "b :\n" << b;
  
   SrMatn x, b2;
   gauss ( a, b, x );

   sr_out << "x :\n" << x;

   b2.mult ( a, x );
   sr_out << "b2 :\n" << b2;

   sr_out << "error : " << dist(b2,b) << srnl;
 }

void test_inverse ( int n )
 {
   SrMatn a(n,n);

   a.set_random ( 0, 99.0 );

   sr_out << "a :\n" << a;

   SrMatn inva(a), id;
   invert ( inva );
   sr_out << "inva :\n" << inva;

   id.mult ( a, inva );
   sr_out << "a * inva :\n" << id;

   SrMatn real_id(n,n);
   real_id.identity();

   sr_out << "error : " << dist(id,real_id) << srnl;
 }


int matn_main ( int, char ** )
 {
   //SrOutputWindow ow ( sr_out );
   test_gauss (5);
   //ow.run();

   return 0;
 }

