
# include <SR/sr_input.h>
# include <SR/sr_string.h>
# include <SR/sr_output.h>
# include <SR/sr_output_window.h>

# include <FL/fl_file_chooser.H>

void parse ( const char *file )
 {
   FILE *stream;
   fopen_s(&stream,file,"r");
   SrInput in ( stream );
   if ( !in.valid() ) sr_out.fatal_error("Could not read file\n");

   SrString buf;
   SrInput::TokenType t;
   while ( !in.finished() )
    { t = in.get_token ( buf );
      const char *s = t==SrInput::Name? "name" :
                t==SrInput::Delimiter? "delim" :
                t==SrInput::Integer? "int" :
                t==SrInput::Real? "real" :
                t==SrInput::String? "string" :
                t==SrInput::EndOfFile? "eof" :
                SrInput::error_desc ( in.last_error() );

      sr_out << "line " << in.curline() << ": " << s << " [" << buf << "]\n";
    }

   sr_out<<srnl;
 }

int parser_main ( int argc, char **argv )
 {
   SrOutputWindow ow ( sr_out );

   if ( argc>=2 ) 
    {
      parse ( argv[1] );
    }
   else
    { char *file = fl_file_chooser ( "Enter File to Read:", "*", 0 );
      if ( file ) parse ( file );
    }

   ow.run();

   return 0;
 }
