
# include <math.h>
# include <SR/sr_output.h>
# include <SR/sr_timer.h>
# include <SR/sr_random.h>

/* There seems to be no time gain from using trigonometric functions based
   on double, float or even integer tables.
   The tests here are even worse for sr_cos() (float versions) because mainly
   of the overhead of having one more function call.
 */

/*! Returns the cosinus of the given integer angle in degrees. This
    function uses a precalculated table, constructed at the first call
    and kept with an internal static pointer. */
float sr_cos ( int ang );

/*! Returns the sinus of the given integer angle in degrees. This
    function uses a precalculated table, constructed at the first call
    and kept with an internal static pointer. */
float sr_sin ( int ang );

static float* cos_table=0;

static void make_cos_table ()
 {
   cos_table = new float[360];
   for ( int i=0; i<360; i++ )
   cos_table[i] = cosf ( SR_TORAD(i) );
 }

float cosi ( int ang )
 {
   if ( !cos_table ) make_cos_table();
   if (ang<0) ang=-ang;
   return cos_table[ ang%360 ];
 }

float sini ( int ang )
 {
   if ( !cos_table ) make_cos_table();
   ang = ang+270;
   if (ang<0) ang=-ang;
   return cos_table[ ang%360 ];
 }

int sincos_main ( int, char ** )
 {
   int i, max=10000;

   SrTimer t;
   SrRandom r;
 
   t.start();
   for ( i=0; i<max; i++ ) cos(SR_TORAD(double(r.get(0,360))));
   t.stop();
   sr_out << "cosd:  " << t.dt() << srnl;

   t.start();
   for ( i=0; i<max; i++ ) cosf(SR_TORAD(float(r.get(0,360))));
   t.stop();
   sr_out << "cosf: " << t.dt() << srnl;

   t.start();
   for ( i=0; i<max; i++ ) cosi(r.get(0,360));
   t.stop();
   sr_out << "cosi: " << t.dt() << srnl;

   int cerrors=0;
   int serrors=0;
   for ( i=-max; i<max; i++ )
    { float s1 = sinf(SR_TORAD(i));
      float s2 = sini(i);
      float c1 = cosf(SR_TORAD(i));
      float c2 = cosi(i);
      if ( SR_DIST(c1,c2)>0.0001f )
       { //sr_out<<c1<<" : "<<c2<<srnl;
         cerrors++;
       }
      if ( SR_DIST(s1,s2)>0.0001f )
       { //sr_out<<s1<<" : "<<s2<<srnl;
         serrors++;
       }
    }

   sr_out << serrors << " sin errors, " << cerrors << " cos errors\n";

   { float x, f, min=0, max=0;
     for ( f=-1.0f; f<=1.0f; f++ )
      { x = acosf(f);
        if ( x<min ) min=x;
        if ( x>max ) max=x;
      }
     sr_out << "acos max="<<max<<" min="<<min<<srnl;
   }

   { float x, f, min=0, max=0;
     for ( f=-1.0f; f<=1.0f; f++ )
      { x = asinf(f);
        if ( x<min ) min=x;
        if ( x>max ) max=x;
      }
     sr_out << "asin max="<<max<<" min="<<min<<srnl;
   }

   { float x, f, min=0, max=0;
     for ( f=-500.0f; f<=500.0f; f++ )
      { x = atanf(f);
        if ( x<min ) min=x;
        if ( x>max ) max=x;
      }
     sr_out << "atan max="<<max<<" min="<<min<<srnl;
   }

   return 0;
 }

