
# include <string.h>

# include <SR/sr_random.h>
# include <SR/sr_string.h>
# include <SR/sr_string_array.h>
# include <SR/sr_array.h>
# include <SR/sr_queue.h>
# include <SR/sr_graph.h>
# include <SR/sr_list.h>
# include <SR/sr_tree.h>
# include <SR/sr_set.h>

class MyListNode : public SrListNode
 { public :
    SrString s;
   public :
    MyListNode () {};
    MyListNode ( const MyListNode& l ) { s=l.s; }
    MyListNode ( const SrString& st ) { s=st; }
   ~MyListNode () {}
    friend SrOutput& operator<< ( SrOutput& o, const MyListNode& l ) { return o<<l.s; }
    friend SrInput& operator>> ( SrInput& i, MyListNode& l ) { return i>>(l.s); }
    friend int sr_compare ( const MyListNode* a, const MyListNode* b ) { return sr_compare(&a->s,&b->s); }
 };

class MyTreeNode : public SrTreeNode
 { public :
    SrString s;
   public :
    MyTreeNode () {}
    MyTreeNode ( const MyTreeNode& n ) { s=n.s; }
    MyTreeNode ( const SrString& st ) { s=st; }
   ~MyTreeNode () {}
    friend SrOutput& operator<< ( SrOutput& o, const MyTreeNode& n ) { return o<<n.s; }
    friend SrInput& operator>> ( SrInput& i, MyTreeNode& n ) { return i>>(n.s); }
    friend int sr_compare ( const MyTreeNode* a, const MyTreeNode* b ) { return sr_compare(&a->s,&b->s); }
 };

class MyGraphNode : public SrGraphNode
 { public :
    SrString s;
   public :
    MyGraphNode () {}
    MyGraphNode ( const MyGraphNode& n ) { s=n.s; }
    MyGraphNode ( const SrString& st ) { s=st; }
   ~MyGraphNode () {}
    friend SrOutput& operator<< ( SrOutput& o, const MyGraphNode& n ) { return o<<n.s; }
    friend SrInput& operator>> ( SrInput& i, MyGraphNode& n ) { return i>>(n.s); }
    friend int sr_compare ( const MyGraphNode* a, const MyGraphNode* b ) { return sr_compare(&a->s,&b->s); }
 };

class MyGraphLink : public SrGraphLink
 { public :
   public :
    friend SrOutput& operator<< ( SrOutput& o, const MyGraphLink& l ) { return o; }
    friend SrInput& operator>> ( SrInput& i, MyGraphLink& l ) { return i; }
    friend int sr_compare ( const MyGraphLink* a, const MyGraphLink* b ) { return 1; }
 };

/* We use this MyBuf class as SrArray and SrQueue cannot be used with classes having
   internal dynamic memory managements. Another option would be to use
   SrArray<SrString*> to have an array of pointers and alloc and delete each entry
   in the array manually. */
struct MyBuf 
 { char buf[20]; 
   void operator= ( const MyBuf& b2 ) { strcpy_s(buf,20,b2.buf); }
   friend SrOutput& operator<< ( SrOutput& o, const MyBuf& b ) { return o<<b.buf; }
   friend int sr_compare ( const MyBuf* a, const MyBuf* b ) { return sr_compare(a->buf,b->buf); }
 };

SrList<MyListNode>    list;
SrTree<MyTreeNode>    tree;
SrGraph<MyGraphNode,MyGraphLink>  graph;
SrSet<SrString>       set;
SrArray<MyBuf>        array;
SrQueue<MyBuf>        queue;
SrStringArray         sarray;

static void out ()
 {
   sr_out << "TREE     : " << tree     <<srnl;
   sr_out << "LIST     : " << list     <<srnl;
   sr_out << "GRAPH    : " << graph    <<srnl;
   sr_out << "SET      : " << set      <<srnl;
   sr_out << "ARRAY    : " << array    <<srnl;
   sr_out << "QUEUE    : " << queue    <<srnl;
   sr_out << "SARRAY   : " << sarray   <<srnl;
 }

int structures_main ( int, char ** )
 {
   int i;
   SrString s;
   SrRandom random(10,99);

   for ( i=0; i<10; i++ ) 
    { int r = random.geti();
      s.setf("A%02d",r); array.push(); strcpy_s(array.top().buf,20,s);
      s.setf("Q%02d",r); queue.insert(); strcpy_s(queue.last().buf,20,s);
      s.setf("S%02d",r); set.insert ( new SrString(s) );
      s.setf("L%02d",r); list.insert_next ( new MyListNode(s) );
      s.setf("G%02d",r); graph.insert ( new MyGraphNode(s) );
      s.setf("T%02d",r); tree.insert_or_del ( new MyTreeNode(s) );
      s.setf("S%02d",r); sarray.push ( s );
    }

   out();

   array.sort(sr_compare);
   sarray.sort();
   list.sort();
   queue.flush();
   queue.compress();
   
   sr_out << "\nAfter sorting arrays and list:\n"; 
   out ();
   return 0;
 }
    
