
# include <SR/sr_vec2.h>
# include <SR/sr_event.h>
# include <SR/sr_plane.h>
# include <SR/sr_camera.h>
# include <SR/sr_sphere.h>
# include <SR/sr_cylinder.h>
# include <SR/sr_sn_group.h>
# include <SR/sr_sn_shape.h>

# include <FL/fl_ask.h>

# include "srxapp_main.h"
# include "srxapp_viewer.h"

# define SR_USE_TRACE1  // event tracing
# include <SR/sr_trace.h>

AppViewer::AppViewer ( int x, int y, int w, int h ) : SrViewer(x,y,w,h) 
 {
   SrSnCylinder* cylinder = new SrSnCylinder;
   cylinder->render_mode ( srRenderModeFlat );
   SrCylinder& c = (SrCylinder&)cylinder->shape();
   c.a.set ( -0.8f, 0, 0 );
   c.b.set ( -0.2f, 0.0f, 0 );
   c.radius = 0.4f;

   SrSnSphere* sphere = new SrSnSphere;
   sphere->render_mode ( srRenderModeFlat );
   sphere->shape().center.set ( 0.5f, 0, 0);
   sphere->shape().radius = 0.3f;

   _root = new SrSnGroup;
   _root->add ( cylinder );
   _root->add ( sphere );

   SrViewer::root ( _root );
   SrViewer::view_all();
 }

float AppViewer::resolution () const
 {
   const SrSnShapeBase* s = (const SrSnShapeBase*) _root->get(0);
   return s->resolution();
 }

void AppViewer::resolution ( float r )
 {
   SrSnShapeBase* s = (SrSnShapeBase*) _root->get(0);
   s->resolution ( r );
   s = (SrSnShapeBase*) _root->get(1);
   s->resolution ( r );
 }

//==================== virtuals ===================================

int AppViewer::handle_keyboard ( const SrEvent& e )
 {
   return SrViewer::handle_keyboard(e);
 }

int AppViewer::handle_scene_event ( const SrEvent& e )
 {
   return SrViewer::handle_scene_event(e);
 }

//============================= EOF ===================================
