
# include <stdlib.h>
# include <stdio.h>

# include <SR/sr_output_window.h>
# include <SR/sr_output.h>

//===== FLTK classes ==================================

# include <FL/Fl.H>
# include <FL/Fl_Window.H>
# include <FL/Fl_Browser.H>
# include <FL/Fl_Menu_Button.H>
# include <FL/fl_ask.H>
# include <FL/fl_file_chooser.H>

enum MenuCmd { CmdSave, CmdActivated, CmdClose, CmdExit };

class OutWin : public Fl_Window
 { public :
    Fl_Menu_Button* _menu;
    OutWin ( int w, int h, const char* title ) : Fl_Window ( w, h, title ) { }
    virtual int handle ( int event );
 };

static void menucb ( Fl_Menu_* o, void* udata ) 
 {
   SrOutputWindow* ow = (SrOutputWindow*) o->parent()->user_data();
   switch ( (MenuCmd)(int)udata )
    { case CmdSave : 
       { char *f = fl_file_chooser ( "Enter Output File:", "*.txt", 0 );
         if ( f ) ow->save(f);
         break;
       }
      case CmdActivated : ow->activated ( !ow->activated() ); break;
      case CmdClose : ow->hide(); break;
      case CmdExit : exit(0); break;
    }
 }

static Fl_Menu_Item MenuTable[] =
 { 
   { "&activate", 0, ((Fl_Callback*)menucb), ((void*)CmdActivated) }, //, FL_MENU_RADIO|FL_MENU_DIVIDER },
   { "&save",     0, ((Fl_Callback*)menucb), ((void*)CmdSave) },
   { "&close",    0, ((Fl_Callback*)menucb), ((void*)CmdClose) },
   { "e&xit all", 0, ((Fl_Callback*)menucb), ((void*)CmdExit) },
   { 0 }
 };

int OutWin::handle ( int event ) 
 {
   bool activate_menu = false;

   if ( event==FL_PUSH )
    if ( Fl::event_state(FL_BUTTON2) || Fl::event_state(FL_BUTTON3) )
     activate_menu=true;

   if ( event==FL_KEYBOARD )
    if ( Fl::event_state(FL_CTRL) && Fl::event_state(FL_SHIFT) && Fl::event_key()=='m' )
     activate_menu=true;

   if ( activate_menu )
    { SrOutputWindow* ow = (SrOutputWindow*) user_data();
      //if ( ow->activated() ) MenuTable[0].set(); else MenuTable[0].clear();
      MenuTable[0].label ( ow->activated()? "de&activate":"&activate" );
      _menu->popup();
      return 1;
    }

   return Fl_Window::handle ( event );
 }

static void make_window ( Fl_Window*& window, Fl_Browser*& browser )
 {
   OutWin* ow = new OutWin(703, 197, "SR Output Window");
   { Fl_Browser* o = new Fl_Browser(1, 0, 702, 197);
     o->textfont(4);
     o->textsize(12);
     Fl_Group::current()->resizable(o);
     browser = o;
   }

   Fl_Menu_Button* m = new Fl_Menu_Button(0,0,50,50);
   m->type(Fl_Menu_Button::POPUP23);
   m->menu(MenuTable);
   m->textsize(12);

   ow->_menu = m;
   ow->end();
   window = ow;
}

//=================== SrOutputWindow =========================

# define GAP  32 // buffer gap

void SrOutputWindow::_init ()
 {
   _window = 0;
   _browser = 0;

   make_window ( _window, _browser );
   _window->user_data ( (void*)(this) ); // to be retreaved by the menu callback

   _buffer=0;
   _bufcur=0;
   _bufcap=0;
   _activated = true;
   _already_shown = false;

   _buffer_len = 512;
 }

SrOutputWindow::SrOutputWindow ()
 {
   _init ();
 }

SrOutputWindow::SrOutputWindow ( SrOutput& o )
 {
   _init ();
   connect ( o ); 
 }

SrOutputWindow::~SrOutputWindow ()
 {
   delete _window;
 }

void SrOutputWindow::save ( const char* fname )
 {
   FILE* fp = fopen ( fname, "wt" );
   if ( !fp ) { fl_alert("Could not write to [%s] !", fname ); return; }

   for ( int i=1; i<=_browser->size(); i++ )
    { fprintf ( fp, "%s\n", _browser->text(i) );
    }

   fclose ( fp );
 }

// we dont use many other SR classes (as SrString) to remain the code
// easy to port to non-SR applications when needed
static void buffer_push ( char*& buf, int& cur, int& cap, char c )
 {
   if ( cur>=cap )
    { cap += GAP;
      buf = (char*) realloc ( buf, cap );
    }
   buf[cur++]=c;
 }

void SrOutputWindow::output ( const char* s )
 {
   if ( !s ) return;
   if ( !_activated ) return;
   if ( !_already_shown ) { _already_shown=true; show(); }

   while ( *s )
    { if ( *s=='\n' )
       { buffer_push(_buffer,_bufcur,_bufcap,0);
         _browser->add ( _buffer );
         _bufcur=0;
         int extra = _browser->size()-_buffer_len;
         if ( extra>0 ) _browser->remove ( 1 );
       }
      else buffer_push(_buffer,_bufcur,_bufcap,*s);
      s++;
    }

   _browser->value ( _browser->size() );
   _browser->deselect();
 }

static void out_func ( const char* s, void* udata )
 {
   ((SrOutputWindow*)udata)->output ( s );
 }

void SrOutputWindow::connect ( SrOutput& o )
 {
   o.init ( out_func, this );
 }

void SrOutputWindow::show ()
 {
   _window->show();
 }

bool SrOutputWindow::shown ()
 {
   return _window->shown()? true:false;
 }

void SrOutputWindow::hide ()
 {
   _window->hide();
 }

void SrOutputWindow::clear ()
 {
   _browser->clear();   
 }

void SrOutputWindow::show_first_line ()
 {
   _browser->value ( 1 );
   _browser->deselect();
   Fl::check ();
 }

void SrOutputWindow::title ( const char* t )
 {
   _window->label(t);
 }

void SrOutputWindow::run ()
 {
   Fl::run();
 }

void SrOutputWindow::check ()
 {
   Fl::check();
 }
