
# include <stdlib.h>
# include <math.h>

# include <FL/Fl.H>
# include <FL/gl.h>
# include <FL/fl_draw.H>
# include <FL/Fl_Menu_Button.H>
# include <FL/Fl_Color_Chooser.H>
# include <FL/Fl_File_Chooser.H>
# include <FL/Fl_Browser.H>

# include <SR/sr_fl.h>
# include <SR/sr_box.h>
# include <SR/sr_quat.h>
# include <SR/sr_line.h>
# include <SR/sr_plane.h>
# include <SR/sr_event.h>
# include <SR/sr_timer.h>
# include <SR/sr_string.h>

# include <SR/sr_gl.h>
# include <SR/sr_light.h>
# include <SR/sr_image.h>
# include <SR/sr_camera.h>
# include <SR/sr_trackball.h>
# include <SR/sr_lines.h>
# include <SR/sr_viewer.h>

# include <SR/sr_sn.h>
# include <SR/sr_sn_group.h>

# include <SR/sr_sa.h>
# include <SR/sr_sa_event.h>
# include <SR/sr_sa_bbox.h>
# include <SR/sr_sa_eps_export.h>
# include <SR/sr_sa_model_export.h>
# include <SR/sr_sa_gl_render.h>

//# define SR_USE_TRACE1  // basic fltk events
//# define SR_USE_TRACE2  // more fltk events
//# define SR_USE_TRACE3  // sr translated events
//# define SR_USE_TRACE4  // view_all
//# define SR_USE_TRACE5  // timeout
# include <SR/sr_trace.h>

//=============================== srSaSetShapesChanged ===========================================

class srSaSetShapesChanged : public SrSa
 { public :
    virtual bool shape_apply ( SrSnShapeBase* s ) { s->changed(true); return true; }
 };

//================================= help window ===================================================

Fl_Window* make_help_window ()
 {
   Fl_Window* win = new Fl_Window ( 300, 200, "SrViewer Help" );
   win->set_non_modal();
   Fl_Browser* b = new Fl_Browser ( 5, 5, 290, 190 );

   b->add ( "Left Click: Select\n" );
   b->add ( "Left Click + Shift: Rotate\n" );
   b->add ( "Left Click + Ctrl: Translate\n" );
   b->add ( "Left Click + Ctrl + Shift: Zoom\n" );
   b->add ( "Left Click + Alt: Y Rotation\n" );
   b->add ( "Ctrl + Shift + m: Menu\n" );
   b->add ( "Ctrl + Shift + x: Exit\n" );
   b->add ( "Ctrl + Shift + c: Print camera\n" );
   b->add ( "Ctrl + Shift + e: EPS export (2D only)\n" );
   b->add ( "Ctrl + Shift + s: Snapshot current frame\n" );
   b->add ( "Ctrl + Shift + a: All frames snapshot on/off\n" );
   b->add ( "Ctrl + Shift + o: Export all models in scene\n" );

   return win;
 }

//================================= popup menu ===================================================

static void menucb ( Fl_Menu_* o, void* v ) 
 {
   ((SrViewer*)(o->parent()->user_data()))->menu_cmd((SrViewer::MenuCmd)(int)v);
 }

# define MCB     ((Fl_Callback*)menucb)
# define CMD(c)  ((void*)SrViewer::c)
static Fl_Menu_Item MenuTable[] =
 { 
   { "&help",       0, MCB, CMD(CmdHelp) },
   { "&view all",   0, MCB, CMD(CmdViewAll) },
   { "&background", 0, MCB, CMD(CmdBackground) }, // FL_MENU_DIVIDER

   { "&mode", 0, 0, 0, FL_SUBMENU },
         { "&examiner", 0, MCB, CMD(CmdExaminer), FL_MENU_RADIO },
         { "&planar",   0, MCB, CMD(CmdPlanar),   FL_MENU_RADIO },
         { 0 },
   { "&draw style", 0, 0, 0, FL_SUBMENU },
         { "&as is",   0, MCB, CMD(CmdAsIs),    FL_MENU_RADIO },
         { "d&efault", 0, MCB, CMD(CmdDefault), FL_MENU_RADIO },
         { "&smooth",  0, MCB, CMD(CmdSmooth),  FL_MENU_RADIO },
         { "&flat",    0, MCB, CMD(CmdFlat),    FL_MENU_RADIO },
         { "&lines",   0, MCB, CMD(CmdLines),   FL_MENU_RADIO },
         { "&points",  0, MCB, CMD(CmdPoints),  FL_MENU_RADIO },
         { 0 },
   { "p&references", 0, 0, 0, FL_SUBMENU },
         { "&axis",         0, MCB, CMD(CmdAxis),        FL_MENU_TOGGLE },
         { "b&ounding box", 0, MCB, CMD(CmdBoundingBox), FL_MENU_TOGGLE },
         { "&statistics",   0, MCB, CMD(CmdStatistics),  FL_MENU_TOGGLE },
         { "spi&n anim",    0, MCB, CMD(CmdSpinAnim),    FL_MENU_TOGGLE },
         { 0 },
   { 0 }
 };
# undef CMD
# undef MCB

// need to set/get data to be able to share the same popup menu with many instances of viewers

static void set_menu_data ( SrViewer::ViewMode v, SrViewer::RenderMode r,
                            bool axis, bool bbox, bool stat, bool spin )
 {
   # define SET(i,b)  if(b) MenuTable[i].set(); else MenuTable[i].clear();
   # define SETO(i)   MenuTable[i].setonly();
   # define CMD(i)    ((int)(MenuTable[i].user_data_))

   int i=0;
   while ( CMD(i)!=SrViewer::CmdExaminer ) i++;      SETO (  i+(int)v );
   while ( CMD(i)!=SrViewer::CmdAsIs ) i++;          SETO (  i+(int)r );
   while ( CMD(i)!=SrViewer::CmdAxis ) i++;          SET  ( i, axis );
   while ( CMD(i)!=SrViewer::CmdBoundingBox ) i++;   SET  ( i, bbox );
   while ( CMD(i)!=SrViewer::CmdStatistics ) i++;    SET  ( i, stat );
   while ( CMD(i)!=SrViewer::CmdSpinAnim ) i++;      SET  ( i, spin );

   # undef CMD
   # undef SETO
   # undef SET
 }

//======================================= timeout ===================================

// The timeout is called by fltk even when the app is iconized.
static void spin_timeout_func ( void* udata ) 
 {
   SrQuat delta;
   double interval;
   double activation;

   //SR_TRACE5 ( "TIMOUT FUNC\n" );

   SrViewer* v = (SrViewer*)udata;
   v->get_spin_data ( delta, interval, activation );

/*   if ( !v->root() && !v->menu_cmd_activated(SrViewer::CmdAxis) )
     v->spinning ( false );
*/
   if ( v->spinning() && !v->iconized() ) 
    { v->increment_model_rotation ( delta );
      v->redraw ();
      Fl::repeat_timeout ( interval, spin_timeout_func, udata );
    }

   v->spin_animation_occured ();
 }

//================================= Internal Structures =================================

struct SpinData
 { SrQuat rotdelta;    // spin current rotation delta used for the spinning animation
   double lasttime;    // last spin time
   double activation;  // spin activation
   double interval;    // spin interval
   SpinData ()  { lasttime=0; init(); }
   void init () { activation=0.1f; interval=0.01f; rotdelta=SrQuat::null; }
   void set_interval ( double i ) { interval = i<0.01? 0.01:i; }
   void set_activation ( double a ) { activation = a<0.01? 0.01:a; }
 };

class SrViewerData
 { public :
   SrSn*  root;              // contains the user scene
   SrViewer::ViewMode viewmode;     // viewer mode, initially Examiner
   SrViewer::RenderMode rendermode; // render mode

   bool iconized;      // to stop processing while the window is iconized
   bool spinning;      // indicates if the model is currently spinning
   bool allowspinanim; // allows spin animation or not
   bool statistics;    // shows statistics or not
   bool displayaxis;   // if shows the axis or not
   bool boundingbox;   // if true will show the bbox of the whole scene
   bool scene_received_event; // to detect and send a release event to the scene graph
                              // when the alt key is released but the mouse is pushed
   SrString message;   // user msg to display in the window
   SrLight light;

   int spin_inc_count;    // counts increments applied to spin animation (for normalization)
   SpinData   spindata;   // Data for spin animation
   SrTrackball trackball; // To process spin rotations
   SrTimer    fcounter;   // To count frames and measure frame rate
   SrEvent    event;      // The translated event from fltk to sr format
   SrColor    bcolor;     // Background color currently used
   SrBox      bbox;       // Bounding box of the root, calculated with viewall
   SrCamera   camera;     // The current camera parameters

   SrSnLines* scenebox;  // contains the bounding box to display, and use in view_all
   SrSnLines* sceneaxis; // the current axis being displayed

   Fl_Menu_Button* menubut; // the ctrl+shift+m or button3 menu
   Fl_Window* helpwin;

   SrSaGlRender render_action;
   SrSaBBox bbox_action;

   SrImage image;
   SrString image_name;
   SrString image_name_full;
   int image_number;
 };

//===================================== SrViewer =================================

// Called when the small "cross" button to close the window is pressed
static void _callback_func ( Fl_Widget* win, void* pt )
 {
   //printf("DBG callback_func!\n");
   SrViewer* v = (SrViewer*)pt;
   v->close_requested ();
 }

SrViewer::SrViewer ( int x, int y, int w, int h, const char *label )
         :Fl_Gl_Window ( x, y, w, h, label )
 {
   Fl::gl_visual ( FL_RGB8  | FL_DOUBLE | FL_DEPTH );//| FL_ALPHA );

   callback ( _callback_func, this );

   resizable(this);

   _data = new SrViewerData;

   _data->root = new SrSnGroup; // we maintain root pointer always valid
   _data->viewmode = ModeExaminer;
   _data->rendermode = ModeAsIs;

   _data->iconized    = false;
   _data->spinning    = false;
   _data->allowspinanim = true;
   _data->statistics  = false;
   _data->displayaxis = false;
   _data->boundingbox = false;
   _data->scene_received_event = false;

   _data->light.init();

   _data->bcolor = SrColor::white;
   _data->spin_inc_count = 0;

   _data->scenebox = new SrSnLines;
   _data->sceneaxis = new SrSnLines;

   _data->image_number = 0; // not saving images

   user_data ( (void*)(this) ); // to be retrieved by the menu callback
   
   begin();
   _data->menubut = new Fl_Menu_Button(0,0,50,50);
   _data->menubut->type(Fl_Menu_Button::POPUP23);
   _data->menubut->menu(MenuTable);
   _data->menubut->textsize(12);
   _data->helpwin = make_help_window ();
   end();
 }

SrViewer::~SrViewer ()
 {
   Fl::remove_timeout ( spin_timeout_func, this );
   _data->root->unref ();
   delete _data->scenebox;
   delete _data->sceneaxis;
   delete _data;
 }

SrSn *SrViewer::root ()
 { 
   return _data->root; 
 }

void SrViewer::root ( SrSn *r )
 { 
   if ( r==_data->root ) return;
   if ( !r ) r = new SrSnGroup;
   _data->root->unref();
   _data->root = r; 
   _data->root->ref();
 }

void SrViewer::draw_message ( const char* s )
 {
   if ( _data->message!=s ) redraw();
   _data->message.set(s);
 }

void SrViewer::show_menu ()
 { 
   set_menu_data ( _data->viewmode, _data->rendermode, _data->displayaxis,
                   _data->boundingbox, _data->statistics, _data->allowspinanim );
   _data->menubut->popup();
 }

void SrViewer::menu_cmd ( MenuCmd s )
 {
   switch ( s )
    { case CmdHelp : _data->helpwin->show(); _data->helpwin->active(); break;

      case CmdViewAll : view_all (); break;

      case CmdBackground :    
       { SrColor c = background();
         if ( fl_color_chooser("Choose background color:",c) ) background ( c );
       } break;

      case CmdExaminer : _data->viewmode = ModeExaminer; 
                         update_axis(); spinning(false); view_all();
                         break;
      case CmdPlanar   : _data->viewmode = ModePlanar;
                         update_axis(); spinning(false); view_all();
                         break;
      case CmdAsIs   : _data->rendermode = ModeAsIs;
                       _data->render_action.restore_render_mode ( _data->root );
                       break;
      case CmdDefault : _data->rendermode = ModeDefault;
                       _data->render_action.override_render_mode ( _data->root, srRenderModeDefault );
                       break;
      case CmdSmooth : _data->rendermode = ModeSmooth;
                       _data->render_action.override_render_mode ( _data->root, srRenderModeSmooth );
                       break;
      case CmdFlat   : _data->rendermode = ModeFlat;
                       _data->render_action.override_render_mode ( _data->root, srRenderModeFlat );
                       break;
      case CmdLines  : _data->rendermode = ModeLines;
                       _data->render_action.override_render_mode ( _data->root, srRenderModeLines );
                       break;
      case CmdPoints : _data->rendermode = ModePoints;
                       _data->render_action.override_render_mode ( _data->root, srRenderModePoints );
                       break;

      case CmdAxis : SR_SWAPB(_data->displayaxis); 
                     if ( _data->displayaxis ) update_axis();
                     break;

      case CmdBoundingBox : SR_SWAPB(_data->boundingbox); 
                            if ( _data->boundingbox ) update_bbox();
                            break;

      case CmdStatistics : SR_SWAPB(_data->statistics); break;
      case CmdSpinAnim   : SR_SWAPB(_data->allowspinanim); 
                           if (!_data->allowspinanim) _data->spinning=false;
                           break;
    }

   render ();
 }

bool SrViewer::menu_cmd_activated ( MenuCmd c )
 {
   switch ( c )
    { case CmdExaminer : return _data->viewmode==ModeExaminer? true:false;
      case CmdPlanar   : return _data->viewmode==ModePlanar? true:false;

      case CmdAsIs     : return _data->rendermode==ModeAsIs? true:false;
      case CmdDefault  : return _data->rendermode==ModeDefault? true:false;
      case CmdSmooth   : return _data->rendermode==ModeSmooth? true:false;
      case CmdFlat     : return _data->rendermode==ModeFlat? true:false;
      case CmdLines    : return _data->rendermode==ModeLines? true:false;
      case CmdPoints   : return _data->rendermode==ModePoints? true:false;

      case CmdAxis        : return _data->displayaxis? true:false;
      case CmdBoundingBox : return _data->boundingbox? true:false;
      case CmdStatistics  : return _data->statistics? true:false;
      case CmdSpinAnim    : return _data->allowspinanim? true:false;
      default : return false;
    }
 }

void SrViewer::update_bbox ()
 {
   _data->bbox_action.apply ( _data->root );
   _data->scenebox->shape().init();
   _data->scenebox->shape().push_box ( _data->bbox_action.get(), true );
 }

void SrViewer::update_axis ()
 {
   _data->bbox_action.apply ( _data->root );
   SrBox b = _data->bbox_action.get();
   float len1 = SR_MAX3(b.a.x,b.a.y,b.a.z);
   float len2 = SR_MAX3(b.b.x,b.b.y,b.b.z);
   float len = SR_MAX(len1,len2);

   _data->sceneaxis->shape().init();

   if ( _data->viewmode==ModePlanar )
    _data->sceneaxis->shape().push_axis ( SrPnt::null, len, 2, "xy" );
   else
    _data->sceneaxis->shape().push_axis ( SrPnt::null, len, 3, "xyz" );
 }

void SrViewer::view_all ()
 {
   _data->spindata.init ();
   _data->spinning = false;

   _data->camera.center = SrVec::null;
   _data->camera.up = SrVec::j;
   _data->camera.eye.set ( 0, 0, 1.0f );

   _data->trackball.init ();

   if ( _data->root )
    { update_bbox ();
      SrBox box = _data->bbox_action.get();

      if ( _data->displayaxis )
       { SrBox b;
         _data->sceneaxis->get_bounding_box(b);
         box.extend ( b );
       }

      float s = box.max_size();
      // _data->light.constant_attenuation = s;
      //_data->camera.view_all ( box, SR_TORAD(60) );
      _data->camera.scale = 1.0f / s;
      //_data->camera.center = box.center();
      //_data->camera.eye = _data->camera.center;
      //_data->camera.eye.z = s.z;
    }

   render ();
 }

void SrViewer::render () 
 { 
   if ( !_data->spinning ) redraw(); 
 } 

bool SrViewer::iconized () 
 { 
   return _data->iconized;
 }

bool SrViewer::spinning ()
 { 
   return _data->spinning;
 }

void SrViewer::set_spin_data ( const SrQuat &delta, float interval, float activation )
 { 
   SpinData &s = _data->spindata;
   s.set_interval ( interval );
   s.rotdelta = delta;
   s.set_activation ( activation );
 }

void SrViewer::get_spin_data ( SrQuat &delta, double &interval, double &activation )
 {
   SpinData &s = _data->spindata;
   delta = s.rotdelta;
   interval = s.interval;
   activation = s.activation;
 }

void SrViewer::spinning ( bool onoff )
 { 
   _data->spinning=onoff;
   if ( _data->spinning )
     Fl::add_timeout ( _data->spindata.interval, spin_timeout_func, (void*)this );
 }

void SrViewer::allow_spin_animation ( bool b )
 { 
   _data->allowspinanim = b;
 }

void SrViewer::increment_model_rotation ( const SrQuat &dq )
 {
   const SrQuat& rotation = _data->trackball.rotation;
   _data->camera *= rotation.inverse() * dq.inverse() * rotation;
   _data->trackball.increment_rotation ( dq );

   if ( _data->spin_inc_count++%300==0 )
    { _data->trackball.rotation.normalize();
      _data->trackball.last_spin.normalize();
      _data->spin_inc_count = 0;
    }
 }

float SrViewer::fps () 
 { 
   return (float)_data->fcounter.mps(); 
 }

sruint SrViewer::curframe () 
 { 
   return (sruint) _data->fcounter.measurements(); 
 }

SrColor SrViewer::background ()
 {
   return _data->bcolor;
 }

void SrViewer::background ( SrColor c )
 {
   _data->bcolor = c;
 }

SrViewer::ViewMode SrViewer::get_view_mode ()
 {
   return _data->viewmode;
 }

void SrViewer::get_camera ( SrCamera &cam )
 {
   cam = _data->camera;
 }

void SrViewer::set_camera ( const SrCamera &cam )
 {
   _data->camera = cam;

   if ( _data->viewmode==ModeExaminer )
    { _data->trackball.init();
      _data->trackball.rotation.set ( cam.eye-cam.center, SrVec::k );
    }
/*
   if ( _data->root )
    { update_bbox ();
      SrBox box = _data->bbox_action.get();
      float s = box.max_size();
      _data->light.constant_attenuation = s;
    }
   else _data->light.constant_attenuation = 1.0f; // the default value
*/
 }

void SrViewer::export_all_models ( const char* prefix, const char* dir )
 {
   SrString pref(prefix);

   if ( !prefix )
    { if ( !fl_string_input ( "Export All Models", "Enter file prefix: ", pref ) )
       return;
    }

   if ( !dir ) 
    { dir = fl_dir_chooser("Enter output directory:",dir);
      if ( !dir ) return;
    }

   SrString s(dir);
   SrSaModelExport exp ( s );
   exp.prefix ( pref );
   exp.apply ( _data->root );
 }

static void gl_draw_string ( const char* s, float x, float y )
 {
   glMatrixMode ( GL_PROJECTION );
   glLoadIdentity ();
   glMatrixMode ( GL_MODELVIEW );
   glLoadIdentity ();
   glDisable ( GL_LIGHTING );
   glColor ( SrColor::red );
   gl_font ( FL_TIMES, 12 ); // from fltk
   gl_draw ( s, x, y );      // from fltk
 }

//-- Render  ------------------------------------------------------------------

void SrViewer::init_opengl ( int w, int h )
 {
   //sr_out<<"INIT"<<srnl;
   glViewport ( 0, 0, w, h );
   glEnable ( GL_DEPTH_TEST );
   glEnable ( GL_LIGHT0 ); 
   glEnable ( GL_LIGHTING );

   //glEnable ( GL_BLEND ); // for transparency
   //glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

   glCullFace ( GL_BACK );
   glDepthFunc ( GL_LEQUAL );
   glFrontFace ( GL_CCW );

   glEnable ( GL_POLYGON_SMOOTH );

   //glEnable ( GL_LINE_SMOOTH );
   //glHint ( GL_LINE_SMOOTH_HINT, GL_NICEST );

   glEnable ( GL_POINT_SMOOTH );
   glPointSize ( 2.0 );

   glShadeModel ( GL_SMOOTH );
 }

void SrViewer::close_requested ()
 {
   exit ( 0 );
 }

/*# include <SR/sr_sphere.h>
static SrSnSphere* SPH=0;
   if ( !SPH ) SPH = new SrSnSphere;
   SPH->shape().center = light.position;
   SPH->shape().radius = 0.6f;
   SPH->color ( SrColor::red );
   _data->render_action.apply ( SPH );
*/   
void SrViewer::draw() 
 {
   if ( !visible() ) return;
   if ( !valid() ) init_opengl ( w(), h() ); // valid() is turned on by fltk after draw() returns
//   sr_out<<"DRAW\n";
 
   SrLight &light = _data->light;
   SrCamera &cam  = _data->camera;
   SrMat mat ( SrMat::NotInitialized );

   light.directional = false;
   light.position = cam.eye;
   light.constant_attenuation = 1.0f/cam.scale;

   //----- Clear Background --------------------------------------------
   glClearColor ( _data->bcolor );
   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   //----- Set Projection ----------------------------------------------
   cam.aspect = (float)w()/(float)h();
   glMatrixMode ( GL_PROJECTION );
   glLoadMatrix ( cam.get_perspective_mat(mat) );

   //----- Set Visualisation -------------------------------------------
   glMatrixMode ( GL_MODELVIEW );
   glLoadMatrix ( cam.get_view_mat(mat) );
   glLight ( 0, light );
   glScalef ( cam.scale, cam.scale, cam.scale );
//   glRotate ( _model_rotation );

   //----- Render user scene -------------------------------------------
   //glClearColor ( _data->bcolor );
   //glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   _data->fcounter.start();
   if ( _data->displayaxis ) _data->render_action.apply ( _data->sceneaxis );
   if ( _data->boundingbox ) _data->render_action.apply ( _data->scenebox );
   if ( _data->root ) _data->render_action.apply ( _data->root );
   _data->fcounter.stop();

   if ( _data->message.len() )
    { gl_draw_string ( _data->message, -1, -1 );
    }
   else if ( _data->statistics )
    { _data->message.setf ( "FPS:%5.2f frame(%2.0f):%4.1fms render:%4.1fms", 
                  _data->fcounter.mps(),
                  _data->fcounter.measurements(),
                  _data->fcounter.loopdt()*1000.0,
                  _data->fcounter.meandt()*1000.0 );
      gl_draw_string ( _data->message, -1.0f, -1.0f );
      _data->message.set ( "" ); // will not deallocate the string but set len==0
    }

   //----- Snapshots -------------------------------------------

   if ( _data->image_number>0 )
    { glSnapshot ( _data->image );
      _data->image_name_full.setf ( "%s%04d.bmp", (const char*)_data->image_name, _data->image_number );
      if ( !_data->image.save_as_bmp(_data->image_name_full) )
        fl_alert("Could not save [%s]!",_data->image_name_full);
      _data->image_number++;
    }

   //----- Fltk will then flush and swap buffers -----------------------------
 }

// Fl::event_x/y() variates from (0,0) to (w(),h())
// transformed coords in SrEvent are in "normalized device coordinates" [-1,-1]x[1,1]
static void translate_event ( SrEvent& e, SrEvent::Type t, int w, int h )
 {
   e.init_lmouse ();

   e.type = t;

   if ( t==SrEvent::Push || t==SrEvent::Release )
    e.button = Fl::event_button();

   if ( Fl::event_state(FL_BUTTON1) ) e.button1 = 1;
   if ( Fl::event_state(FL_BUTTON2) ) e.button2 = 1;
   if ( Fl::event_state(FL_BUTTON3) ) e.button3 = 1;

   if ( Fl::event_state(FL_ALT)   ) e.alt = 1;
   if ( Fl::event_state(FL_CTRL)  ) e.ctrl = 1;

   if ( Fl::event_state(FL_SHIFT) ) e.shift = 1;
   
   e.key = Fl::event_key();

   // put coordinates inside [-1,1] with (0,0) in the middle :
   e.mouse.x  = ((float)Fl::event_x())*2.0f / ((float)w) - 1.0f;
   e.mouse.y  = ((float)Fl::event_y())*2.0f / ((float)h) - 1.0f;
   e.mouse.y *= -1.0f;
   e.width = w;
   e.heigth = h;
 }

int SrViewer::handle ( int event ) 
 {
   # define POPUP_MENU(e) e.button3

   SrEvent &e = _data->event;
   e.type = SrEvent::None;

   switch ( event )
    { case FL_PUSH :
       { SR_TRACE1 ( "Mouse Push : but="<<Fl::event_button()<<" ("<<Fl::event_x()<<", "<<Fl::event_y()<<")" <<" Ctrl:"<<Fl::event_state(FL_CTRL) );
         translate_event ( e, SrEvent::Push, w(), h() );
         if ( POPUP_MENU(e) ) { show_menu(); e.type=SrEvent::None; }
          else _data->spinning=false; 
       } break;

      case FL_RELEASE:
        SR_TRACE1 ( "Mouse Release : ("<<Fl::event_x()<<", "<<Fl::event_y()<<") buts: "
                     <<(Fl::event_state(FL_BUTTON1)?1:0)<<" "<<(Fl::event_state(FL_BUTTON2)?1:0) );
        translate_event ( e, SrEvent::Release, w(), h() );
        break;

      case FL_MOVE:
        SR_TRACE2 ( "Move buts: "<<(Fl::event_state(FL_BUTTON1)?1:0)<<" "<<(Fl::event_state(FL_BUTTON2)?1:0) );
        if ( !Fl::event_state(FL_BUTTON1) && !Fl::event_state(FL_BUTTON2) ) break;
        // otherwise, this is a drag: enter in the drag case.
        // not sure if this is a hack or a feature.
      case FL_DRAG:
        SR_TRACE2 ( "Mouse Drag : ("<<Fl::event_x()<<", "<<Fl::event_y()<<") buts: "
                     <<(Fl::event_state(FL_BUTTON1)?1:0)<<" "<<(Fl::event_state(FL_BUTTON2)?1:0) );
        translate_event ( e, SrEvent::Drag, w(), h() );
        break;

      case FL_SHORTCUT: // not sure the relationship between a shortcut and keyboard event...
        SR_TRACE1 ( "Shortcut : "<< Fl::event_key() <<" "<<Fl::event_text() );
        translate_event ( e, SrEvent::Keyboard, w(), h() );
        break;
      
      case FL_KEYBOARD:
        SR_TRACE1 ( "Key Pressed : "<< Fl::event_key() <<" "<<Fl::event_text() );
//        translate_event ( e, SrEvent::Keyboard, w(), h() );
        break;

      case FL_HIDE: // Called when the window is iconized
        { SR_TRACE1 ( "Hide" );
          _data->iconized = true;
          // the opengl lists need to be re-created when the window appears again, so
          // we mark already here all shapes as changed:
          _data->scenebox->changed(true);
          _data->sceneaxis->changed(true);
          srSaSetShapesChanged sa;
          sa.apply ( _data->root );
        } break;

      case FL_SHOW: // Called when the window is de-iconized or when show() is called
        SR_TRACE1 ( "Show" );
        _data->iconized = false;
        if ( _data->spinning ) Fl::add_timeout ( _data->spindata.interval, spin_timeout_func, (void*)this );
        show ();
        break;

      // Other events :
      case FL_ENTER:          SR_TRACE2 ( "Enter" );         break;
      case FL_LEAVE:          SR_TRACE2 ( "Leave" );         break;
      case FL_FOCUS:          SR_TRACE2 ( "Focus" );         break;
      case FL_UNFOCUS:        SR_TRACE2 ( "Unfocus" );       break;
      case FL_CLOSE:          SR_TRACE2 ( "Close");          break;
      case FL_ACTIVATE:       SR_TRACE2 ( "Activate");       break;
      case FL_DEACTIVATE:     SR_TRACE2 ( "Deactivate");     break;
      case FL_PASTE:          SR_TRACE2 ( "Paste");          break;
      case FL_SELECTIONCLEAR: SR_TRACE2 ( "SelectionClear"); break;
    }

   SR_TRACE3 ( e );

   if ( e.type == SrEvent::None ) return 0; // not an interesting event

   if ( event==FL_PUSH || event==FL_DRAG )
    { SrPlane plane ( _data->camera.center, SrVec::k );
      _data->camera.get_ray ( e.mouse.x, e.mouse.y, e.ray.p1, e.ray.p2 );
      _data->camera.get_ray ( e.lmouse.x, e.lmouse.y, e.lray.p1, e.lray.p2 );
      e.mousep = plane.intersect ( e.ray.p1, e.ray.p2 );
      e.lmousep = plane.intersect ( e.lray.p1, e.lray.p2 );
      if ( event==FL_PUSH  ) // update picking precision
       { // define a and b with 1 pixel difference:
         SrPnt2 a ( ((float)w())/2.0f, ((float)h())/2.0f ); // ( float(Fl::event_x()), float(Fl::event_y()) );
         SrPnt2 b (a+SrVec2::one);// ( float(Fl::event_x()+1), float(Fl::event_y()+1) );
         // put coordinates inside [-1,1] with (0,0) in the middle :
         a.x  = a.x*2.0f / float(w()) - 1.0f;
         a.y  = a.y*2.0f / float(h()) - 1.0f; a.y *= -1.0f;
         b.x  = b.x*2.0f / float(w()) - 1.0f;
         b.y  = b.y*2.0f / float(h()) - 1.0f; b.y *= -1.0f;
         //sr_out << "a,b: " << a << srspc << b <<srnl;
         SrLine aray, bray;
         _data->camera.get_ray ( a.x, a.y, aray.p1, aray.p2 );
         _data->camera.get_ray ( b.x, b.y, bray.p1, bray.p2 );
         SrPnt pa = plane.intersect ( aray.p1, aray.p2 );
         SrPnt pb = plane.intersect ( bray.p1, bray.p2 );
         //sr_out << "pa,pb: " << pa << srspc << pb <<srnl;
         e.pixel_size = (SR_DIST(pa.x,pb.x)+SR_DIST(pa.y,pb.y))/2.0f;
         //sr_out << "pixel_size: " << e.pixel_size <<srnl;
       }
    }

   return handle_event ( e );
 }

//== handle sr event =======================================================

int SrViewer::handle_event ( const SrEvent &e )
 {
   int res=0;

   if ( (e.ctrl || e.shift || e.alt) && e.mouse_event() )
    { 
      if ( _data->viewmode==ModeExaminer )
        res = handle_examiner_manipulation ( e );
      else if ( _data->viewmode==ModePlanar )
        res = handle_planar_manipulation ( e );

      if ( res ) return res;
    }

   if ( e.mouse_event() ) return handle_scene_event ( e );

   if ( e.type == SrEvent::Keyboard )
    { if ( handle_keyboard(e)==0 ) res = handle_scene_event ( e );
      if ( res==0 && e.key==SrEvent::KeyEsc ) res=1; // to avoid exiting with ESC
    }

   return res; // this point should not be reached
 }

//== Examiner ==============================================================

# define ROTATING(e)    (e.shift&&e.button1)
# define ROTATING2(e)   (e.alt&&e.button1)
# define ZOOMING(e)     (e.shift&&e.ctrl&&e.button1) || (e.shift&&e.button1&&e.button2)
# define TRANSLATING(e) (e.ctrl&&e.button1) || (e.shift&&e.button2)

int SrViewer::handle_examiner_manipulation ( const SrEvent &e )
 {
   if ( e.type==SrEvent::Drag )
    { 
      float dx = e.mousedx() * _data->camera.aspect;
      float dy = e.mousedy() / _data->camera.aspect;

      if ( ZOOMING(e) )
       { _data->camera.fovy += (dx+dy);//*40.0f;
         _data->camera.fovy = SR_BOUND ( _data->camera.fovy, 0.001f, srpi );
       }
      else if ( TRANSLATING(e) )
       { _data->camera.apply_translation_from_mouse_motion ( e.lmouse.x, e.lmouse.y, e.mouse.x, e.mouse.y );
       }
      else if ( ROTATING(e) )
       { SrQuat drot;
         SrTrackball::get_spin_from_mouse_motion ( e.lmouse.x, e.lmouse.y, e.mouse.x, e.mouse.y, drot );
         increment_model_rotation ( drot );
        _data->spindata.lasttime = _data->fcounter.time();
       }
      else if ( ROTATING2(e) )
       { SrQuat q ( SrVec::j, -(e.mouse.x-e.lmouse.x)*2.0f ); // around Y
         _data->camera.center = _data->camera.center*q;
         _data->camera.eye = _data->camera.eye * q;
         _data->camera.up = _data->camera.up * q;
       }
      redraw();
    }
   else if ( e.type==SrEvent::Release )
    { if ( e.button==1 && _data->allowspinanim && _data->spindata.lasttime>0 ) 
       { _data->spindata.set_interval ( _data->fcounter.time()-_data->spindata.lasttime );
         _data->spindata.rotdelta = _data->trackball.last_spin;
         SR_TRACE5 ( _data->spindata.interval<<" < "<<_data->spindata.activation );
         if ( _data->spindata.interval<_data->spindata.activation )
          { _data->spinning=true;
            Fl::add_timeout ( _data->spindata.interval, spin_timeout_func, (void*)this );
          }
       }
    }
   return 1;
 }

//== Planar =============================================================

int SrViewer::handle_planar_manipulation ( const SrEvent& e )
 {
   SrCamera& c = _data->camera;

   if ( e.type==SrEvent::Drag )
    { 
      float dx = e.mousedx() * c.aspect;
      float dy = e.mousedy() / c.aspect;

      if ( ZOOMING(e) ) // scaling effect in planar mode
       { c.fovy += (dx+dy)*2;
         c.fovy = SR_BOUND ( c.fovy, 0.01f, srpi );
       }
      else if ( TRANSLATING(e) ) // this will translate in planar mode
       { _data->camera.apply_translation_from_mouse_motion ( e.lmouse.x, e.lmouse.y, e.mouse.x, e.mouse.y );
       }
      else if ( ROTATING(e) ) // planar mode
       { 
         SrQuat drot;
         SrPnt p1 ( e.lmousep.x, e.lmousep.y, 0 );
         SrPnt p2 ( e.mousep.x, e.mousep.y, 0 );
         drot.set ( p1, p2 );
         c.up = c.up * drot;
       }

      redraw();
    }

   return 1;
 }

//== Apply Scene action ==========================================================

int SrViewer::handle_scene_event ( const SrEvent& e )
 {
   SrSaEvent ea(e);
   ea.apply ( _data->root );
   int used = ea.result();
   if ( used ) render();
   return used;
 }

//== Keyboard ==============================================================

static void eps_export ( SrSn* n )
 { 
   char *file = fl_file_chooser ( "Export to .eps file", "*.eps", 0 );
   if ( !file ) return;
   SrOutput o ( fopen(file,"wt") );
   if ( !o.valid() ) return;
   SrSaEpsExport eps ( o );
   eps.apply ( n );
 }

static void snapshot ()
 {
   char *file = fl_file_chooser ( "Capture to .bmp image", "*.bmp", 0 );
   if ( !file ) return;
   SrImage image;
   glSnapshot ( image );
   if ( !image.save_as_bmp(file) ) fl_alert("Could not save [%s]!",file);
 }

static void frame_snapshot_onoff ( SrViewerData* d )
 {
   if ( d->image_number>0 ) // turn off
    { int n = d->image_number-1;
      d->image_number=0;
      d->image.alloc(0,0);
      fl_message ( "Frame Snapshot Turned Off.\n%d images saved.", n );
    }
   else // turn on
    { char *file = fl_file_chooser ( "Enter base .bmp file", "*.bmp", 0 );
      if ( !file ) return;
      d->image_name = file;
      if ( d->image_name.has_file_extension("bmp") )
       d->image_name.remove_file_extension();
      if ( d->image_name.last_char()=='.' ) d->image_name.last_char(0);
      d->image_number = 1;
    }
 }

static void print_camera ( SrCamera& c )
 { 
   SrString buf;
   SrOutput o(buf);
   o << c;
   fl_message ( buf );
 }

int SrViewer::handle_keyboard ( const SrEvent &e )
 {
   if ( e.ctrl && e.shift )
    { switch ( e.key )
       { case 'm' : show_menu (); return 1;
         case 'x' : hide (); return 1;
         case 'c' : print_camera ( _data->camera); return 1;
         case 'e' : eps_export ( _data->root ); return 1;
         case 's' : snapshot (); return 1;
         case 'a' : frame_snapshot_onoff ( _data ); return 1;
         case 'o' : export_all_models(); return 1;
       }
    }
   return 0;
 }

//== Spin Animation ========================================================

void SrViewer::spin_animation_occured ()
 {
 }

//================================ End of File =================================================
