 
# include <stdlib.h>

# include <FL/Fl.H>
# include <FL/fl_file_chooser.H>

# include <SR/sr_event.h>
# include <SR/sr_camera.h>
# include <SR/sr_box.h>
# include <SR/sr_model.h>
# include <SR/sr_lines.h>
# include <SR/sr_sn_shape.h>
# include <SR/sr_sn_group.h>
# include <SR/sr_sn_matrix.h>
# include <SR/sr_viewer.h>
# include <SR/sr_output.h>

class CamViewer;
class ObjViewer;

CamViewer *TheCamViewer=0;
ObjViewer *TheObjViewer=0;

//=========================== CamViewer =============================

class CamViewer : public SrViewer
 { public :
    SrSnLines *Ray;
    SrSnLines *Cam;
   public :
    CamViewer ( SrSn* n, int x, int y, int w, int h );
 };

CamViewer::CamViewer ( SrSn* n, int x, int y, int w, int h ) : SrViewer(x,y,w,h)
 {
   Cam = new SrSnLines;
   Ray = new SrSnLines;

   SrSnGroup* g = new SrSnGroup;
   g->add ( Cam );
   g->add ( Ray );
   g->add ( n );
   root ( g );
 }

//=========================== ObjViewer =============================

class ObjViewer : public SrViewer
 { public :
    ObjViewer ( SrSn* n, int x, int y, int w, int h );

    virtual int handle_event ( const SrEvent& e );
    virtual int handle_scene_event ( const SrEvent& e );
    virtual void spin_animation_occured ();
 };

ObjViewer::ObjViewer ( SrSn* n, int x, int y, int w, int h ) : SrViewer(x,y,w,h)
 {
   menu_cmd ( SrViewer::CmdAxis );

   root ( n );
   view_all();
 }

void ObjViewer::spin_animation_occured ()
 {
   SrEvent e;
   handle_event ( e ); // just to update the cam window while spinning
 }

int ObjViewer::handle_event ( const SrEvent& e )
 {
   SrCamera cam;
   get_camera ( cam );
   TheCamViewer->Cam->shape().init ();
   TheCamViewer->Cam->shape().push_color ( SrColor::black );
   TheCamViewer->Cam->shape().push_line ( cam.eye, cam.center );
   TheCamViewer->Cam->shape().push_line ( cam.eye, cam.eye+(cam.up*0.1f) );
   TheCamViewer->render();

   if ( e.type==SrEvent::Keyboard )
    { 
      if ( e.key==SrEvent::KeyEsc ) exit(0);

      if ( e.key==SrEvent::KeySpace )
       { get_camera ( cam );
         set_camera ( cam );
         render();
         return 1;
       }
    }

   return SrViewer::handle_event(e);
 }

int ObjViewer::handle_scene_event ( const SrEvent& e )
 {
   SrVec p1, p2;
   SrCamera cam;
   get_camera ( cam );
   cam.get_ray ( e.mouse.x, e.mouse.y, p1, p2 );

   TheCamViewer->Ray->shape().init ();
   TheCamViewer->Ray->shape().push_color ( SrColor::cyan );
   TheCamViewer->Ray->shape().push_line ( p1, p2 );

   return 1; // tell that I understood the event
 }

//=========================== main =============================

int camera_main ( int, char ** )
 {
   SrSnModel* model = new SrSnModel;

   SrInput in ( fopen("../data/knot.srm","rt") );

   if ( !model->shape().load(in) )
    { char *s = 0;
      while ( !s )
       { s = fl_file_chooser ( "Enter model", "*.srm", 0 );
         if ( !s ) continue;
         SrInput in ( fopen(s,"rt") );
         if ( !model->shape().load(in) ) s=0;
       }
    }

   SrMat mat;
   // light is not working properly. I tried to put it also at the same
   // position of the eye (see draw() in sr_scene_viewer.cpp), but is
   // not working, to see the problem try to put 0.5 or 2.0 in the scale.
   mat.scale ( 1.0f );

   // problems arise also when we apply a translation...
   mat.setl4 ( 0.0f, 0, 0, 1.0f );

   SrSnGroup* g = new SrSnGroup;
   g->add ( model );

   SrSnMatrix* m = new SrSnMatrix;
   m->set ( mat );
   g->add ( m );

   TheObjViewer = new ObjViewer ( g, 385, 50, 350, 280 );
   TheObjViewer->menu_cmd ( SrViewer::CmdAxis );
   TheObjViewer->show();

   TheCamViewer = new CamViewer ( g, 30, 50, 350, 280 );
   TheCamViewer->show();
   TheCamViewer->menu_cmd ( SrViewer::CmdAxis );

   Fl::run();

   return 0;
 }
